import socket
from logging import config
from typing import List

import aiohttp
import asyncio
import pippin.config as config
import nanopy
import os
import rapidjson as json

from pippin.db.redis import RedisDB
from pippin.network.dpow_websocket import ConnectionClosed, DpowClient
from pippin.util.nano_util import NanoUtil

class WorkClient(object):
    _instance = None

    def __init__(self):
        raise RuntimeError('Call instance() instead')

    @classmethod
    def instance(cls) -> 'WorkClient':
        if cls._instance is None:
            cls._instance = cls.__new__(cls)
            cls.work_urls = config.Config.instance().work_peers
            if config.Config.instance().node_work_generate:
                cls.work_urls.append(config.Config.instance().node_url)                
            cls.session = aiohttp.ClientSession(json_serialize=json.dumps)
            cls.dpow_client = None
            cls.dpow_futures = {}
            cls.dpow_id = 1
            # Construct DPoW Client
            cls.dpow_user = os.getenv('DPOW_USER', None)
            cls.dpow_key = os.getenv('DPOW_KEY', None)
            if cls.dpow_user is not None and cls.dpow_key is not None:
                cls.dpow_client = DpowClient(
                    cls.dpow_user,
                    cls.dpow_key,
                    work_futures=cls.dpow_futures,
                    bpow=False
                )
                cls.dpow_fallback_url = 'https://dpow.nanocenter.org/service/'
            else:
                cls.dpow_user = os.getenv('BPOW_USER', None)
                cls.dpow_key = os.getenv('BPOW_KEY', None)
                if cls.dpow_user is not None and cls.dpow_key is not None:
                    cls.dpow_client = DpowClient(
                        cls.dpow_user,
                        cls.dpow_key,
                        work_futures=cls.dpow_futures,
                        bpow=True
                    )
                    cls.dpow_fallback_url = 'https://bpow.banano.cc/service/'

        return cls._instance

    @classmethod
    async def close(cls):
        if hasattr(cls, 'session') and cls.session is not None:
            await cls.session.close()
        if cls._instance is not None:
            cls._instance = None

    async def make_request(self, url: str, req_json: dict):
        async with self.session.post(url ,json=req_json, timeout=300) as resp:
            return await resp.json()

    async def work_generate(self, hash: str, difficulty: str) -> str:
        work_generate = {
            'action': 'work_generate',
            'hash': hash,
            'difficulty': difficulty
        }

        # Build work_generate requests
        tasks = []
        for p in self.work_urls:
            tasks.append(self.make_request(p, work_generate))

        # Use DPoW if applicable
        if self.dpow_client is not None:
            dpow_id = str(self.dpow_id)
            self.dpow_id += 1
            self.dpow_futures[dpow_id] = asyncio.get_event_loop().create_future()
            try:
                success = await self.dpow_client.request_work(dpow_id, hash, difficulty=difficulty)
                tasks.append(self.dpow_futures[dpow_id])
            except ConnectionClosed:
                # HTTP fallback for this request
                dp_req = {
                    "user": self.dpow_user,
                    "api_key": self.dpow_key,
                    "hash": hash,
                    "difficulty": difficulty
                }
                tasks.append(self.make_request(self.dpow_fallback_url, dp_req))

        # Do it locally if no peers or if peers have been failing
        if await RedisDB.instance().exists("work_failure") or (len(self.work_urls) == 0 and self.dpow_client is None):
            tasks.append(
                NanoUtil.instance().work_generate(hash, difficulty=difficulty)
            )

        # Post work_generate to all peers simultaneously
        final_result = None
        while len(tasks) > 0:
            # Fire all tasks simultaneously and fire when first one is completed
            done, pending = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED, timeout=100)
            # Set to True if any tasks completed before the timeout
            has_done = len(done) > 0
            try:
                for task in done:
                    result = task.result()
                    # Some tasks return different types of responses, e.g. DPoW is different than a normal work peer
                    if result is None:
                        aiohttp.log.server_logger.info("work_generate task returned None")
                    if isinstance(result, list):
                        result = json.loads(result[1])
                    elif isinstance(result, str):
                        result = {'work':result}
                    if result is not None and 'work' in result:
                        cancel_json = {
                            'action': 'work_cancel',
                            'hash': hash
                        }
                        for p in self.work_urls:
                            asyncio.ensure_future(self.make_request(p, cancel_json))
                        final_result = result['work']
                    elif result is not None and 'error' in result:
                        aiohttp.log.server_logger.info(f'work_generate task returned error {result["error"]}')
            except Exception:
                aiohttp.log.server_logger.exception("work_generate task raised Exception")
            finally:
                # Either nothing finished before the timeout or something did
                # Cancel any pending tasks
                if not has_done or final_result is not None or len(pending) == 0:
                    for p in pending:
                        try:
                            p.cancel()
                        except Exception:
                            pass
                    break
                elif len(pending) > 0:
                    # wait again for pending tasks
                    tasks = pending

        if final_result is not None:
            return final_result

        # IF we're still here then all requests failed, set failure flag
        await RedisDB.instance().set(f"work_failure", "aa", expires=300)
        return await NanoUtil.instance().work_generate(hash, difficulty=difficulty)



import json
import os
import pathlib

from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop
from tortoise import Tortoise

from pippin.config import Config
from pippin.db.models.wallet import Wallet
from pippin.db.tortoise_config import DBConfig
from pippin.db.redis import RedisDB
from pippin.server.pippin_server import PippinServer
from pippin.util.utils import Utils


class PippinServerTest(AioHTTPTestCase):
    async def get_application(self):
        self.server = PippinServer('127.0.0.1', 9999)
        return self.server.app

    @classmethod
    def setUpClass(cls):
        return super().setUpClass()

    async def setUpAsync(self):
        if 'BANANO' in os.environ:
            del os.environ['BANANO']
        await DBConfig(mock=True).init_db()
        await RedisDB.instance().get_redis()

    async def tearDownAsync(self):
        if 'BANANO' in os.environ:
            del os.environ['BANANO']
        await Tortoise.close_connections()
        await RedisDB.close()
        try:
            os.remove(Utils.get_project_root().joinpath(pathlib.PurePath('mock.db')))
        except FileNotFoundError:
            pass
        try:
            os.remove(Utils.get_project_root().joinpath(pathlib.PurePath('mock.db-wal')))
        except FileNotFoundError:
            pass
        try:
            os.remove(Utils.get_project_root().joinpath(pathlib.PurePath('mock.db-shm')))
        except FileNotFoundError:
            pass

    async def json_request(self, data: str):
        return await self.client.request(
            'POST',
            '/',
            data=data,
            headers={
                'Content-Type': 'application/json'
            }
        )

    async def create_test_wallet(self):
        body = json.dumps(
            {
                'action': 'wallet_create',
                'seed':'C273AB6E1D8121C5DA0B99DD44CF9AA29D51C40B009ACB9410CA1649E28170E8'
            }
        )
        response = await self.json_request(body)
        json_resp = await response.json()
        return json_resp['wallet']


    @unittest_run_loop
    async def test_bad_request(self):
        body = json.dumps(
            {
                'sdasd': 'asdadas'
            }
        )
        response = await self.json_request(body)
        json_resp = await response.json()
        self.assertTrue('error' in json_resp)

    @unittest_run_loop
    async def test_wallet_create(self):
        body = json.dumps(
            {
                'action': 'wallet_create',
            }
        )
        response = await self.json_request(body)
        self.assertEqual(response.status, 200)

        json_resp = await response.json()
        self.assertTrue('wallet' in json_resp)

    @unittest_run_loop
    async def test_wallet_create_with_seed(self):
        wallet_id = await self.create_test_wallet()

        # Make sure we can get wallet
        wallet: Wallet = await Wallet.get_wallet(wallet_id)
        self.assertEqual(wallet.seed, 'C273AB6E1D8121C5DA0B99DD44CF9AA29D51C40B009ACB9410CA1649E28170E8')
        account = await wallet.get_account('nano_3n5jcd764a1t5duoc455zhrz6ernn534c9onrao4tghhfpw8y6m65og66qtc')
        self.assertEqual(account.address, 'nano_3n5jcd764a1t5duoc455zhrz6ernn534c9onrao4tghhfpw8y6m65og66qtc')
        self.assertEqual(account.account_index, 0)

    @unittest_run_loop
    async def test_account_create(self):
        wallet_id = await self.create_test_wallet()

        body = json.dumps(
            {
                'action': 'account_create',
                'wallet':wallet_id
            }
        )
        response = await self.json_request(body)
        self.assertEqual(response.status, 200)
        json_resp = await response.json()
        self.assertTrue('account' in json_resp)
        # Equal to index 1 account
        self.assertEqual(json_resp['account'], 'nano_3yfyxm9aeyawqwfcd6c6zd36ph877xwwpxeqppn5wp6rzeifrnpi1p49fhr9')

    @unittest_run_loop
    async def test_accounts_create(self):
        wallet_id = await self.create_test_wallet()

        body = json.dumps(
            {
                'action': 'accounts_create',
                'wallet':wallet_id,
                'count':5
            }
        )
        response = await self.json_request(body)
        self.assertEqual(response.status, 200)
        json_resp = await response.json()
        self.assertTrue('accounts' in json_resp)
        # Equal to index 1-5 accounts
        self.assertEqual(json_resp['accounts'][0], 'nano_3yfyxm9aeyawqwfcd6c6zd36ph877xwwpxeqppn5wp6rzeifrnpi1p49fhr9')
        self.assertEqual(json_resp['accounts'][1], 'nano_1pdqbxoe81uzdd3u7ft91u49ztdkgdgfoeak7jmysbia7msigbctunmq9php')
        self.assertEqual(json_resp['accounts'][2], 'nano_3a34k3bgc147r6emhac6jxgzquqif695hi8xqypowdru3eum81je3no4iwty')
        self.assertEqual(json_resp['accounts'][3], 'nano_3im7mxo6717mhc54t9ypyber981y8perupjap5p9ysf4wymf5e4a64ugnjax')
        self.assertEqual(json_resp['accounts'][4], 'nano_3cptr3ispmjte8ky1e15imt6py7yppwp1h4r7kfyxy1z3xjbjeujmca4g3m1')

    @unittest_run_loop
    async def test_account_list(self):
        wallet_id = await self.create_test_wallet()
        # Create accounts
        body = json.dumps(
            {
                'action': 'accounts_create',
                'wallet':wallet_id,
                'count':5
            }
        )
        response = await self.json_request(body)
        self.assertEqual(response.status, 200)

        # Test list
        body = json.dumps(
            {
                'action': 'account_list',
                'wallet':wallet_id
            }
        )
        response = await self.json_request(body)
        self.assertEqual(response.status, 200)
        json_resp = await response.json()
        self.assertTrue('accounts' in json_resp)
        # Equal to index 0-5 accounts
        self.assertEqual(json_resp['accounts'][0], 'nano_3n5jcd764a1t5duoc455zhrz6ernn534c9onrao4tghhfpw8y6m65og66qtc')
        self.assertEqual(json_resp['accounts'][1], 'nano_3yfyxm9aeyawqwfcd6c6zd36ph877xwwpxeqppn5wp6rzeifrnpi1p49fhr9')
        self.assertEqual(json_resp['accounts'][2], 'nano_1pdqbxoe81uzdd3u7ft91u49ztdkgdgfoeak7jmysbia7msigbctunmq9php')
        self.assertEqual(json_resp['accounts'][3], 'nano_3a34k3bgc147r6emhac6jxgzquqif695hi8xqypowdru3eum81je3no4iwty')
        self.assertEqual(json_resp['accounts'][4], 'nano_3im7mxo6717mhc54t9ypyber981y8perupjap5p9ysf4wymf5e4a64ugnjax')
        self.assertEqual(json_resp['accounts'][5], 'nano_3cptr3ispmjte8ky1e15imt6py7yppwp1h4r7kfyxy1z3xjbjeujmca4g3m1')

    @unittest_run_loop
    async def test_account_list(self):
        wallet_id = await self.create_test_wallet()
        # Create accounts
        body = json.dumps(
            {
                'action': 'accounts_create',
                'wallet':wallet_id,
                'count':5
            }
        )
        response = await self.json_request(body)
        self.assertEqual(response.status, 200)

        # Test list
        body = json.dumps(
            {
                'action': 'account_list',
                'wallet':wallet_id
            }
        )
        response = await self.json_request(body)
        self.assertEqual(response.status, 200)
        json_resp = await response.json()
        self.assertTrue('accounts' in json_resp)
        # Equal to index 0-5 accounts
        self.assertEqual(json_resp['accounts'][0], 'nano_3n5jcd764a1t5duoc455zhrz6ernn534c9onrao4tghhfpw8y6m65og66qtc')
        self.assertEqual(json_resp['accounts'][1], 'nano_3yfyxm9aeyawqwfcd6c6zd36ph877xwwpxeqppn5wp6rzeifrnpi1p49fhr9')
        self.assertEqual(json_resp['accounts'][2], 'nano_1pdqbxoe81uzdd3u7ft91u49ztdkgdgfoeak7jmysbia7msigbctunmq9php')
        self.assertEqual(json_resp['accounts'][3], 'nano_3a34k3bgc147r6emhac6jxgzquqif695hi8xqypowdru3eum81je3no4iwty')
        self.assertEqual(json_resp['accounts'][4], 'nano_3im7mxo6717mhc54t9ypyber981y8perupjap5p9ysf4wymf5e4a64ugnjax')
        self.assertEqual(json_resp['accounts'][5], 'nano_3cptr3ispmjte8ky1e15imt6py7yppwp1h4r7kfyxy1z3xjbjeujmca4g3m1')

    @unittest_run_loop
    async def test_password_change(self):
        wallet_id = await self.create_test_wallet()
        # Set password
        body = json.dumps(
            {
                'action': 'password_change',
                'wallet':wallet_id,
                'password': 'abcd123'
            }
        )
        response = await self.json_request(body)
        self.assertEqual(response.status, 200)
        json_resp = await response.json()
        self.assertTrue('changed' in json_resp)
        self.assertEqual(json_resp['changed'], '1')

        # Ensure we get wallet locked error when changing again
        # Set password
        body = json.dumps(
            {
                'action': 'password_change',
                'wallet':wallet_id,
                'password': 'def567'
            }
        )
        response = await self.json_request(body)
        self.assertEqual(response.status, 200)
        json_resp = await response.json()
        self.assertTrue('error' in json_resp)
        self.assertEqual(json_resp['error'].lower(), 'wallet locked')

    @unittest_run_loop
    async def test_password_enter(self):
        wallet_id = await self.create_test_wallet()
        # Set password
        body = json.dumps(
            {
                'action': 'password_change',
                'wallet':wallet_id,
                'password': 'abcd123'
            }
        )
        response = await self.json_request(body)
        self.assertEqual(response.status, 200)
        json_resp = await response.json()
        self.assertTrue('changed' in json_resp)
        self.assertEqual(json_resp['changed'], '1')

        # Ensure bad password doesn't unlock
        body = json.dumps(
            {
                'action': 'password_enter',
                'wallet':wallet_id,
                'password': 'def567'
            }
        )
        response = await self.json_request(body)
        self.assertEqual(response.status, 200)
        json_resp = await response.json()
        self.assertTrue('valid' in json_resp)
        self.assertEqual(json_resp['valid'], '0')
        body = json.dumps(
            {
                'action': 'password_change',
                'wallet':wallet_id,
                'password': 'def567'
            }
        )
        response = await self.json_request(body)
        self.assertEqual(response.status, 200)
        json_resp = await response.json()
        self.assertTrue('error' in json_resp)
        self.assertEqual(json_resp['error'].lower(), 'wallet locked')

        # Test good password
        body = json.dumps(
            {
                'action': 'password_enter',
                'wallet':wallet_id,
                'password': 'abcd123'
            }
        )
        response = await self.json_request(body)
        self.assertEqual(response.status, 200)
        json_resp = await response.json()
        self.assertTrue('valid' in json_resp)
        self.assertEqual(json_resp['valid'], '1')
        body = json.dumps(
            {
                'action': 'password_change',
                'wallet':wallet_id,
                'password': 'def567'
            }
        )
        response = await self.json_request(body)
        self.assertEqual(response.status, 200)
        json_resp = await response.json()
        self.assertTrue('changed' in json_resp)
        self.assertEqual(json_resp['changed'], '1')

    @unittest_run_loop
    async def test_wallet_lock_locked(self):
        wallet_id = await self.create_test_wallet()
        # Set password
        body = json.dumps(
            {
                'action': 'password_change',
                'wallet':wallet_id,
                'password': 'abcd123'
            }
        )
        response = await self.json_request(body)
        self.assertEqual(response.status, 200)
        json_resp = await response.json()
        self.assertTrue('changed' in json_resp)
        self.assertEqual(json_resp['changed'], '1')

        # Check locked stats
        body = json.dumps(
            {
                'action': 'wallet_locked',
                'wallet':wallet_id
            }
        )
        response = await self.json_request(body)
        self.assertEqual(response.status, 200)
        json_resp = await response.json()
        self.assertTrue('locked' in json_resp)
        self.assertEqual(json_resp['locked'], '1') # Locked

        # Unlock
        body = json.dumps(
            {
                'action': 'password_enter',
                'wallet':wallet_id,
                'password': 'abcd123'
            }
        )
        response = await self.json_request(body)
        self.assertEqual(response.status, 200)
        json_resp = await response.json()
        self.assertTrue('valid' in json_resp)
        self.assertEqual(json_resp['valid'], '1')

        # Check locked stats
        body = json.dumps(
            {
                'action': 'wallet_locked',
                'wallet':wallet_id
            }
        )
        response = await self.json_request(body)
        self.assertEqual(response.status, 200)
        json_resp = await response.json()
        self.assertTrue('locked' in json_resp)
        self.assertEqual(json_resp['locked'], '0') # Unlocked

        # Lock
        body = json.dumps(
            {
                'action': 'wallet_lock',
                'wallet':wallet_id
            }
        )
        response = await self.json_request(body)
        self.assertEqual(response.status, 200)
        json_resp = await response.json()
        self.assertTrue('locked' in json_resp)
        self.assertEqual(json_resp['locked'], '1')

        # Check locked stats
        body = json.dumps(
            {
                'action': 'wallet_locked',
                'wallet':wallet_id
            }
        )
        response = await self.json_request(body)
        self.assertEqual(response.status, 200)
        json_resp = await response.json()
        self.assertTrue('locked' in json_resp)
        self.assertEqual(json_resp['locked'], '1') # Locked

    @unittest_run_loop
    async def test_wallet_representative_set(self):
        wallet_id = await self.create_test_wallet()
        # Set rep
        body = json.dumps(
            {
                'action': 'wallet_representative_set',
                'wallet':wallet_id,
                'representative': 'nano_1he11o1darknessmyo1dfriend11111111111111111111111111ki7nzfx6'
            }
        )
        response = await self.json_request(body)
        self.assertEqual(response.status, 200)
        json_resp = await response.json()
        self.assertTrue('set' in json_resp)
        self.assertEqual(json_resp['set'], '1')

        # Ensure we get right representative when querying
        body = json.dumps(
            {
                'action': 'wallet_representative',
                'wallet':wallet_id
            }
        )
        response = await self.json_request(body)
        self.assertEqual(response.status, 200)
        json_resp = await response.json()
        self.assertTrue('representative' in json_resp)
        self.assertEqual(json_resp['representative'], 'nano_1he11o1darknessmyo1dfriend11111111111111111111111111ki7nzfx6')

        # Ensure we can't set a bad rep
        body = json.dumps(
            {
                'action': 'wallet_representative_set',
                'wallet':wallet_id,
                'representative': 'nano_1he11o1darknessmyo1dfriend11111111111111111111111111ki7nzfx5'
            }
        )
        response = await self.json_request(body)
        self.assertEqual(response.status, 200)
        json_resp = await response.json()
        self.assertTrue('error' in json_resp)
        self.assertEqual(json_resp['error'].lower(), 'invalid address')

    @unittest_run_loop
    async def test_adhoc_account(self):
        wallet_id = await self.create_test_wallet()
        # Set rep
        body = json.dumps(
            {
                'action': 'wallet_add',
                'wallet':wallet_id,
                'key': '11A780CD68BF64AE703C9AA7C138E5D6F917EBB60C55D50BE8BD6AFBA65066F8'
            }
        )
        response = await self.json_request(body)
        self.assertEqual(response.status, 200)
        json_resp = await response.json()
        self.assertTrue('account' in json_resp)
        self.assertEqual(json_resp['account'], 'nano_1997zr1zatrca8giqmyyz45a7q8qqgqkxdntg8uashgrqe5odmz93q5u954d')

    @unittest_run_loop
    async def test_wallet_change_seed(self):
        wallet_id = await self.create_test_wallet()
        # Change seed
        body = json.dumps(
            {
                'action': 'wallet_change_seed',
                'wallet':wallet_id,
                'seed': 'D8BF40997C092631640856E2CD47FCA1BB39FE678ED34665580E8FF61FA1C049'
            }
        )
        response = await self.json_request(body)
        self.assertEqual(response.status, 200)
        json_resp = await response.json()
        self.assertTrue('success' in json_resp)
        self.assertTrue('last_restored_account' in json_resp)
        self.assertTrue('restored_count' in json_resp)
        self.assertEqual(json_resp['last_restored_account'], 'nano_1997zr1zatrca8giqmyyz45a7q8qqgqkxdntg8uashgrqe5odmz93q5u954d')
        self.assertEqual(json_resp['restored_count'], 1)



from tortoise.models import Model
from tortoise import fields

import nanopy

class Account(Model):
    wallet  = fields.ForeignKeyField('db.Wallet', on_delete=fields.CASCADE, related_name='accounts', index=True)
    address = fields.CharField(max_length=65)
    account_index = fields.IntField()
    work = fields.BooleanField(default=True)
    created_at = fields.DatetimeField(auto_now_add=True)

    class Meta:
        table = 'accounts'
        unique_together = ('address', 'account_index')

    def private_key(self, seed: str) -> str:
        """Return account private key in hex"""
        return nanopy.deterministic_key(seed, self.account_index)[0]


import logging
import os
from tortoise import Tortoise
from pippin.util.utils import Utils
import pathlib

class DBConfig(object):
    def __init__(self, mock = False):
        self.logger = logging.getLogger()
        self.modules = {'db': ['pippin.db.models.wallet', 'pippin.db.models.account', 'pippin.db.models.adhoc_account', 'pippin.db.models.block']}
        self.mock = mock
        if self.mock:
            self.use_postgres = False
            self.use_mysql = False
            return
        # Postgres
        self.use_postgres = False
        self.postgres_db = os.getenv('POSTGRES_DB')
        self.postgres_user = os.getenv('POSTGRES_USER')
        self.postgres_password = os.getenv('POSTGRES_PASSWORD')
        self.postgres_host = os.getenv('POSTGRES_HOST', '127.0.0.1')
        self.postgres_port = os.getenv('POSTGRES_PORT', 5432)
        if self.postgres_db is not None and self.postgres_user is not None and self.postgres_password is not None:
            self.use_postgres = True
        elif self.postgres_db is not None or self.postgres_user is not None or self.postgres_password is not None:
            raise Exception("ERROR: Postgres is not properly configured. POSTGRES_DB, POSTGRES_USER, and POSTGRES_PASSWORD environment variables are all required.")
        # MySQL
        self.use_mysql = False
        if not self.use_postgres:
            self.mysql_db = os.getenv('MYSQL_DB')
            self.mysql_user = os.getenv('MYSQL_USER')
            self.mysql_password = os.getenv('MYSQL_PASSWORD')
            self.mysql_host = os.getenv('MYSQL_HOST', '127.0.0.1')
            self.mysql_port = os.getenv('MYSQL_PORT', 3306)
            if self.mysql_db is not None and self.mysql_user is not None and self.mysql_password is not None:
                self.use_mysql = True
            elif self.mysql_db is not None or self.mysql_user is not None or self.mysql_password is not None:
                raise Exception("ERROR: Postgres is not properly configured. MYSQL_DB, MYSQL_USER, and MYSQL_PASSWORD environment variables are all required.")

    async def init_db(self):
        if self.use_postgres:
            self.logger.info(f"Using PostgreSQL Database {self.postgres_db}")
            await Tortoise.init(
                db_url=f'postgres://{self.postgres_user}:{self.postgres_password}@{self.postgres_host}:{self.postgres_port}/{self.postgres_db}',
                modules=self.modules
            )
        elif self.use_mysql:
            self.logger.info(f"Using MySQL Database {self.mysql_db}")
            await Tortoise.init(
                db_url=f'mysql://{self.mysql_user}:{self.mysql_password}@{self.mysql_host}:{self.mysql_port}/{self.mysql_db}',
                modules=self.modules
            )
        else:
            self.logger.info(f"Using SQLite database pippin.db")
            dbpath = Utils.get_project_root().joinpath(pathlib.PurePath('pippin.db')) if not self.mock else Utils.get_project_root().joinpath(pathlib.PurePath('mock.db'))
            await Tortoise.init(
                db_url=f'sqlite://{dbpath}',
                modules=self.modules
            )
        # Create tables
        await Tortoise.generate_schemas(safe=True)


from pippin.util.crypt import AESCrypt, DecryptionError

import unittest

class TestAESCrypt(unittest.TestCase):
    def setUp(self):
        self.crypt1 = AESCrypt('mypassword')
        self.crypt2 = AESCrypt('someotherpassword')

    def test_encrypt_decrypt(self):
        """Test encryption and decryption"""
        some_string = "Some Random String to Encrypt"
        encrypted = self.crypt1.encrypt(some_string)
        self.assertNotEqual(encrypted, some_string)
        self.assertEqual(self.crypt1.decrypt(encrypted), some_string)
        with self.assertRaises(DecryptionError) as exc:
            self.crypt2.decrypt(encrypted)


from pippin.util.random import RandomUtil

import unittest

class TestRandomUtil(unittest.TestCase):
    def test_seed_generate(self):
        """Test seed generation"""
        self.assertTrue(len(RandomUtil.generate_seed()) == 64)
        for i in range(5):
            self.assertNotEqual(RandomUtil.generate_seed(), RandomUtil.generate_seed())


from pippin.util.wallet import WalletUtil
from pippin.util.nano_util import NanoUtil
from pippin.util.utils import Utils
from tortoise import Tortoise
from pippin.db.tortoise_config import DBConfig
from pippin.db.redis import RedisDB
from pippin.db.models.account import Account
from pippin.db.models.adhoc_account import AdHocAccount
from pippin.db.models.wallet import Wallet

import os
import asyncio
import unittest
import uuid
import pathlib

class TestWalletUtil(unittest.TestCase):
    @classmethod
    async def setUpClassAsync(cls):
        if 'BANANO' in os.environ:
            del os.environ['BANANO']
        await DBConfig(mock=True).init_db()
        cls.wallet = Wallet(
            seed="7474F694061FB3E5813986AEC8A65340B5DEDB4DF94E394CB44489BEA6B21FCD",
            representative='nano_1oa4sipdm679m7emx9npmoua14etkrj1e85g3ujkxmn4ti1aifbu7gx1zrgr',
            encrypted=False
        )
        await cls.wallet.save()
        cls.account = Account(
            wallet=cls.wallet,
            address='nano_19zdjp6tfhqzcag9y3w499h36nr16ks6gdsfkawzgomrbxs54xaybmsyamza',
            account_index=0
        )
        await cls.account.save()
        cls.adhoc_account = AdHocAccount(
            wallet=cls.wallet,
            address='nano_1oa4sipdm679m7emx9npmoua14etkrj1e85g3ujkxmn4ti1aifbu7gx1zrgr',
            private_key='86A3D926AB6BEBAA678C13823D7A92A97CAFAFD277EBF4B54C42C8BB9806EAEE'
        )
        await cls.adhoc_account.save()
        cls.wallet_util = WalletUtil(cls.account, cls.wallet)
        cls.wallet_util_adhoc = WalletUtil(cls.adhoc_account, cls.wallet)

    @classmethod
    def removeMockDB(cls):
        try:
            os.remove(Utils.get_project_root().joinpath(pathlib.PurePath('mock.db')))
        except FileNotFoundError:
            print(str(Utils.get_project_root().joinpath(pathlib.PurePath('mock.db'))))
            pass
        try:
            os.remove(Utils.get_project_root().joinpath(pathlib.PurePath('mock.db-wal')))
        except FileNotFoundError:
            pass
        try:
            os.remove(Utils.get_project_root().joinpath(pathlib.PurePath('mock.db-shm')))
        except FileNotFoundError:
            pass

    @classmethod
    def setUpClass(cls):
        cls.removeMockDB()
        cls.loop = asyncio.new_event_loop()
        cls.loop.run_until_complete(cls.setUpClassAsync())

    @classmethod
    def tearDownClass(cls):
        cls.loop.run_until_complete(Tortoise.close_connections())
        cls.loop.run_until_complete(NanoUtil.close())
        cls.loop.close()
        cls.removeMockDB()

    def test_get_representative(self):
        self.assertEqual(self.wallet_util.get_representative(), 'nano_1oa4sipdm679m7emx9npmoua14etkrj1e85g3ujkxmn4ti1aifbu7gx1zrgr')

    def test_adhoc(self):
        self.assertFalse(self.wallet_util.adhoc())
        self.assertTrue(self.wallet_util_adhoc.adhoc())

    def test_private_key_derivation(self):
        """Ensure  private keys can be derived correctly"""
        self.assertEqual('DD21B99F1A92D7315BF8592F632B05EC77555DFD6A0B2D99C560537CAFC9A13E', self.wallet_util.private_key().upper())
        self.assertEqual('86A3D926AB6BEBAA678C13823D7A92A97CAFAFD277EBF4B54C42C8BB9806EAEE', self.wallet_util_adhoc.private_key().upper())




from typing import TYPE_CHECKING, List, Union

import aioredis
import nanopy
import rapidjson
from aiohttp import log
from aioredlock import LockError

import pippin.config as config
import pippin.util.nano_util as nano_util
from pippin.db.redis import RedisDB
from pippin.db.models.account import Account
from pippin.db.models.adhoc_account import AdHocAccount
from pippin.db.models.block import Block
from pippin.model.difficulty import DifficultyModel
from pippin.network.rpc_client import AccountNotFound, RPCClient
from pippin.network.work_client import WorkClient

if TYPE_CHECKING:
    from pippin.db.models.wallet import Wallet

class WalletUtil(object):
    """Wallet utilities, like signing, creating blocks, etc."""

    def __init__(self, acct: Union[Account, AdHocAccount], wallet: 'Wallet'):
        self.account = acct
        self.wallet = wallet

    def get_representative(self):
        if self.wallet.representative is None:
            return config.Config.instance().get_random_rep()
        return self.wallet.representative

    def adhoc(self) -> bool:
        return not isinstance(self.account, Account)

    def private_key(self) -> str:
        if self.adhoc():
            return self.account.private_key_get()
        return self.account.private_key(self.wallet.seed)


    async def publish(self, state_block: dict, subtype: str = None) -> dict:
        """Publish a state block"""
        try:
            resp = await RPCClient.instance().process(state_block, subtype=subtype)
            # The RPC send/receive uses `block` as a key instead of `hash`
            if 'hash' in resp:
                return {'block': resp['hash']}
            return resp
        except Exception:
            from pippin.db.models.wallet import ProcessFailed
            raise ProcessFailed()

    async def _receive_block_create(self, hash: str, work: str = None) -> dict:
        """Build a state block (receive)"""
        # Get block info
        block_info = await RPCClient.instance().block_info(hash)
        if block_info is None or block_info['contents']['link_as_account'] != self.account.address:
            return None
        # Get account info
        is_open = True
        try:
            account_info = await RPCClient.instance().account_info(self.account.address)
            if account_info is None:
                return None
        except AccountNotFound:
            is_open = False

        # Different workbase for open/receive
        if is_open:
            workbase = account_info['frontier']
        else:
            workbase = nanopy.account_key(self.account.address)

        # Build other fields
        previous = '0000000000000000000000000000000000000000000000000000000000000000' if not is_open else account_info['frontier']
        representative = self.get_representative() if not is_open else account_info['representative']
        balance = block_info['amount'] if not is_open else str(int(account_info['balance']) + int(block_info['amount']))

        # Generate work
        if work is None:
            try:
                work = await WorkClient.instance().work_generate(workbase, DifficultyModel.instance().receive_difficulty)
                if work is None:
                    log.server_logger.error("WORK FAILED")
                    raise WorkFailed(workbase)
            except Exception:
                log.server_logger.exception("work failed")
                raise WorkFailed(workbase)

        # Build final state block
        state_block = nanopy.state_block()
        state_block['account'] = self.account.address
        state_block['previous'] = previous
        state_block['representative'] = representative
        state_block['balance'] = balance
        state_block['link'] = hash
        state_block['work'] = work

        # Sign block
        pk = self.private_key()
        state_block['signature'] = await nano_util.NanoUtil.instance().sign_block(pk, block=state_block)

        return state_block

    async def receive(self, hash: str, work: str = None) -> dict:
        """Receive a block and return hash of published block"""
        async with await (await RedisDB.instance().get_lock_manager()).lock(f"pippin:{self.account.address}") as lock:
            block = await self._receive_block_create(hash, work)
            return await self.publish(block, subtype='receive')

    async def _receive_all(self) -> int:
        """Receive and publish multiple blocks"""
        received_count = 0
        p = await RPCClient.instance().pending(self.account.address, threshold=config.Config.instance().receive_minimum)
        if p is None:
            return received_count
        for block in p:
            state_block = await self._receive_block_create(block)
            await self.publish(state_block, subtype='receive')
            received_count += 1
        return received_count

    async def receive_all(self) -> int:
        """Receive all pending blocks for this account and return # received"""
        received_count = 0
        async with await (await RedisDB.instance().get_lock_manager()).lock(f"pippin:{self.account.address}") as lock:
            received_count = await self._receive_all()
        return received_count

    async def _send_block_create(self, amount: int, destination: str, id: str = None, work: str = None) -> dict:
        """Create a state block (send)"""
        # Get account info
        account_balance = await RPCClient.instance().account_balance(self.account.address)
        if account_balance is None:
            return None

        # Check balance
        if amount > int(account_balance['balance']):
            # Auto-receive blocks if they have it pending
            if config.Config.instance().auto_receive_on_send and int(account_balance['balance']) + int(account_balance['pending']) >= amount:
                await self._receive_all()
                account_info = await RPCClient.instance().account_info(self.account.address)
                if account_info is None:
                    return None
                if amount > int(account_info['balance']):
                    raise InsufficientBalance(account_info['balance'])
            else:
                raise InsufficientBalance(account_balance['balance'])

        # Get account info
        account_info = await RPCClient.instance().account_info(self.account.address)
        if account_info is None:
            return None
        
        workbase = account_info['frontier']

        # Build other fields
        previous = account_info['frontier']
        representative = account_info['representative']
        balance = str(int(account_info['balance']) - amount)

        # Generate work
        if work is None:
            try:
                work = await WorkClient.instance().work_generate(workbase, DifficultyModel.instance().send_difficulty)
                if work is None:
                    raise WorkFailed(workbase)
            except Exception:
                raise WorkFailed(workbase)

        # Build final state block
        state_block = nanopy.state_block()
        state_block['account'] = self.account.address
        state_block['previous'] = previous
        state_block['representative'] = representative
        state_block['balance'] = balance
        state_block['link'] = nanopy.account_key(destination.replace("xrb_", "nano_"))
        state_block['work'] = work

        # Sign block
        pk = self.private_key()
        state_block['signature'] = await nano_util.NanoUtil.instance().sign_block(pk, block=state_block)

        return state_block

    async def send(self, amount: int, destination: str, id: str = None, work: str = None) -> dict:
        """Create a send block and return hash of published block
            amount is in RAW"""
        
        async with await (await RedisDB.instance().get_lock_manager()).lock(f"pippin:{self.account.address}") as lock:
            # See if block exists, if ID specified
            # If so just rebroadcast it and return the hash
            if id is not None:
                if not self.adhoc():
                    block = await Block.filter(send_id=id, account=self.account).first()
                else:
                    block = await Block.filter(send_id=id, adhoc_account=self.account).first()
                if block is not None:
                    await RPCClient.instance().process(block.block)
                    return {"block": block.block_hash.upper()}
            # Create block
            state_block = await self._send_block_create(amount, destination, id=id, work=work)
            # Publish block
            resp = await self.publish(state_block, subtype='send')
            # Cache if ID specified
            if resp is not None and 'block' in resp:
                # Cache block in database if it has id specified
                if id is not None:
                    block = Block(
                        account=self.account if not self.adhoc() else None,
                        adhoc_account=self.account if self.adhoc() else None,
                        block_hash=nanopy.block_hash(state_block),
                        block=state_block,
                        send_id=id,
                        subtype='send'
                    )
                    await block.save()
            return resp

    async def _change_block_create(self, representative: str, work: str = None, only_if_different: bool = False) -> dict:
        """Create a state block (change)"""
        # Get account info
        account_info = await RPCClient.instance().account_info(self.account.address)
        if account_info is None:
            return None
        elif only_if_different and account_info['representative'] == representative:
            return None

        workbase = account_info['frontier']

        # Build other fields
        previous = account_info['frontier']
        representative = representative
        balance = account_info['balance']

        # Generate work
        if work is None:
            try:
                work = await WorkClient.instance().work_generate(workbase, DifficultyModel.instance().send_difficulty)
                if work is None:
                    raise WorkFailed(workbase)
            except Exception:
                raise WorkFailed(workbase)

        # Build final state block
        state_block = nanopy.state_block()
        state_block['account'] = self.account.address
        state_block['previous'] = previous
        state_block['representative'] = representative
        state_block['balance'] = balance
        state_block['link'] = '0000000000000000000000000000000000000000000000000000000000000000'
        state_block['work'] = work

        # Sign block
        pk = self.private_key()
        state_block['signature'] = await nano_util.NanoUtil.instance().sign_block(pk, block=state_block)

        return state_block

    async def representative_set(self, representative: str, work: str = None, only_if_different: bool = False) -> dict:
        """Create a change block and return hash of published block"""
        async with await (await RedisDB.instance().get_lock_manager()).lock(f"pippin:{self.account.address}") as lock:
            state_block = await self._change_block_create(representative, work=work, only_if_different=only_if_different)

            if state_block is None and only_if_different:
                return None

            # Publish
            return await self.publish(state_block, subtype='change')

class WorkFailed(Exception):
    pass

class ProcessFailed(Exception):
    pass

class InsufficientBalance(Exception):
    pass



from tortoise.models import Model
from tortoise import fields
from pippin.model.secrets import SeedStorage

class AdHocAccount(Model):
    wallet  = fields.ForeignKeyField('db.Wallet', on_delete=fields.CASCADE, related_name='adhoc_accounts', index=True)
    address = fields.CharField(max_length=65)
    private_key = fields.CharField(max_length=128)
    work = fields.BooleanField(default=True)
    created_at = fields.DatetimeField(auto_now_add=True)

    class Meta:
        table = 'adhoc_accounts'
        unique_together = ('address', 'private_key')

    def private_key_get(self) -> str:
        """Return account private key in hex"""
        priv_key = SeedStorage.instance().get_decrypted_seed(f"{self.wallet_id}:{self.address}")
        return priv_key if priv_key is not None else self.private_key


import base64
import hashlib

from Crypto import Random
from Crypto.Cipher import AES

class AESCrypt():

    def __init__(self, secret_key: str):
        self.key = hashlib.sha256(secret_key.encode('utf-8')).digest()
        self.BS = 16
        self.salt = '61606982'

    pad = lambda self, s: s + (self.BS - len(s) % self.BS) * chr(self.BS - len(s) % self.BS)
    unpad = lambda self, s: s[0:-s[-1]]

    def encrypt(self, value: str) -> str:
        value = self.salt + value
        value = self.pad(value)
        iv = Random.new().read(AES.block_size)
        cipher = AES.new(self.key, AES.MODE_CBC, iv)
        return base64.b64encode(iv + cipher.encrypt(value.encode('latin-1'))).decode('latin-1')

    def decrypt(self, encrypted: str) -> str:
        encrypted = base64.b64decode(encrypted.encode('latin-1'))
        iv = encrypted[:16]
        cipher = AES.new(self.key, AES.MODE_CBC, iv)
        decrypted = self.unpad(cipher.decrypt(encrypted[16:])).decode('latin-1')
        # Test if decrypt was successful
        if len(decrypted) < len(self.salt):
            raise DecryptionError()
        elif decrypted[:len(self.salt)] != self.salt:
            raise DecryptionError()
        return decrypted[len(self.salt):]

class DecryptionError(Exception):
    pass



import pathlib
from dotenv import load_dotenv
load_dotenv()
from pippin.util.utils import Utils
load_dotenv(dotenv_path=Utils.get_project_root().joinpath(pathlib.PurePath('.env')))

import argparse
import asyncio
import getpass

from pippin.db.models.wallet import Wallet, WalletLocked, WalletNotFound
from pippin.db.tortoise_config import DBConfig
from tortoise import Tortoise
from tortoise.transactions import in_transaction
from pippin.util.crypt import AESCrypt, DecryptionError
from pippin.util.random import RandomUtil
from pippin.util.validators import Validators
from pippin.version import __version__

from pippin.config import Config
import os

# Set and patch nanopy
import nanopy
nanopy.account_prefix = 'ban_' if Config.instance().banano else 'nano_'
if Config.instance().banano:
    nanopy.standard_exponent = 29
    nanopy.work_difficulty = 'fffffe0000000000'

parser = argparse.ArgumentParser(description=f'Pippin v{__version__}')
subparsers = parser.add_subparsers(title='available commands', dest='command')

wallet_parser = subparsers.add_parser('wallet_list')

wallet_create_parser = subparsers.add_parser('wallet_create')
wallet_create_parser.add_argument('--seed', type=str, help='Seed for wallet (optional)', required=False)

wallet_change_seed_parser = subparsers.add_parser('wallet_change_seed')
wallet_change_seed_parser.add_argument('--wallet', type=str, help='ID of wallet to change seed for', required=True)
wallet_change_seed_parser.add_argument('--seed', type=str, help='New seed for wallet (optional)', required=False)
wallet_change_seed_parser.add_argument('--encrypt', action='store_true', help='If specified, will get prompted for a password to encrypt the wallet', default=False)

wallet_view_seed_parser = subparsers.add_parser('wallet_view_seed')
wallet_view_seed_parser.add_argument('--wallet', type=str, help='Wallet ID', required=True)
wallet_view_seed_parser.add_argument('--password', type=str, help='Password needed to decrypt wallet (if encrypted)', required=False)
wallet_view_seed_parser.add_argument('--all-keys', action='store_true', help='Also show all of the wallet address and keys', default=False)

account_create_parser = subparsers.add_parser('account_create')
account_create_parser.add_argument('--wallet', type=str, help='Wallet ID', required=True)
account_create_parser.add_argument('--key', type=str, help='AdHoc Account Key', required=False)
account_create_parser.add_argument('--count', type=int, help='Number of accounts to create (min: 1)', required=False)

wallet_destroy_parser = subparsers.add_parser('wallet_destroy')
wallet_destroy_parser.add_argument('--wallet', type=str, help='Wallet ID', required=True)

repget_parser = subparsers.add_parser('wallet_representative_get')
repget_parser.add_argument('--wallet', type=str, help='Wallet ID', required=True)

repset_parser = subparsers.add_parser('wallet_representative_set')
repset_parser.add_argument('--wallet', type=str, help='Wallet ID', required=True)
repset_parser.add_argument('--representative', type=str, help='New Wallet Representative', required=True)
repset_parser.add_argument('--update-existing', action='store_true', help='Update existing accounts', default=False)

options = parser.parse_args()

async def wallet_list():
    wallets = await Wallet.all().prefetch_related('accounts', 'adhoc_accounts')
    if len(wallets) == 0:
        print("There aren't any wallets")
        return

    for w in wallets:
        print(f"ID:{w.id}")
        print("Accounts:")
        for a in w.accounts:
            print(a.address)

async def wallet_create(seed):
    async with in_transaction() as conn:
        wallet = Wallet(
            seed=RandomUtil.generate_seed() if seed is None else seed
        )
        await wallet.save(using_db=conn)
        new_acct = await wallet.account_create(using_db=conn)
    print(f"Wallet created, ID: {wallet.id}\nFirst account: {new_acct}")

async def wallet_change_seed(wallet_id: str, seed: str, password: str) -> str:
    encrypt = False
    old_password = None
    if len(password) > 0:
        encrypt = True

    # Retrieve wallet
    try:
        wallet = await Wallet.get_wallet(wallet_id)
    except WalletNotFound:
        print(f"No wallet found with ID: {wallet_id}")
        exit(1)
    except WalletLocked as wl:
        wallet = wl.wallet
        while True:
            try:
                npass = getpass.getpass(prompt='Enter current password:')
                crypt = AESCrypt(npass)
                try:
                    decrypted = crypt.decrypt(wallet.seed)
                    async with in_transaction() as conn:
                        wallet.seed = decrypted
                        wallet.encrypted = False
                        await wallet.save(using_db=conn, update_fields=['seed', 'encrypted'])
                        for a in await wallet.adhoc_accounts.all():
                            a.private_key = crypt.decrypt(a.private_key)
                            await a.save(using_db=conn, update_fields=['private_key'])
                    old_password = npass
                    break
                except DecryptionError:
                    print("**Invalid password**")
            except KeyboardInterrupt:
                break
                exit(0)

    # Change key
    await wallet.change_seed(seed)

    # Encrypt if necessary
    if encrypt:
        await wallet.encrypt_wallet(password)

    # Get newest account
    newest = await wallet.get_newest_account()

    print(f"Seed changed for wallet {wallet.id}\nFirst account: {newest.address}")

async def wallet_view_seed(wallet_id: str, password: str, all_keys: bool) -> str:
    # Retrieve wallet
    crypt = None
    try:
        wallet = await Wallet.get_wallet(wallet_id)
    except WalletNotFound:
        print(f"No wallet found with ID: {wallet_id}")
        exit(1)
    except WalletLocked as wl:
        wallet = None
        if password is not None:
            crypt = AESCrypt(password)
            try:
                decrypted = crypt.decrypt(wl.wallet.seed)
                wallet = wl.wallet
                wallet.seed = decrypted
            except DecryptionError:
                pass
        if wallet is None:
            while True:
                try:
                    npass = getpass.getpass(prompt='Enter current password:')
                    crypt = AESCrypt(npass)
                    try:
                        decrypted = crypt.decrypt(wl.wallet.seed)
                        wallet = wl.wallet
                        wallet.seed = decrypted
                    except DecryptionError:
                        print("**Invalid password**")
                except KeyboardInterrupt:
                    break
                    exit(0)

    print(f"Seed: {wallet.seed}")
    if all_keys:
        for a in await wallet.accounts.all():
            print(f"Addr: {a.address} PrivKey: {nanopy.deterministic_key(wallet.seed, index=a.account_index)[0].upper()}")
    else:
        print(f"AdHoc accounts:")
        for a in await wallet.adhoc_accounts.all():
            if not wallet.encrypted:
                print(f"Addr: {a.address} PrivKey: {a.private_key.upper()}")
            else:
                print(f"Addr: {a.address} PrivKey: {crypt.decrypt(a.private_key)}")

async def account_create(wallet_id: str, key: str, count: int = 1) -> str:
    # Retrieve wallet
    crypt = None
    password=None
    if count is None:
        count = 1
    try:
        wallet = await Wallet.get_wallet(wallet_id)
    except WalletNotFound:
        print(f"No wallet found with ID: {wallet_id}")
        exit(1)
    except WalletLocked as wl:
        wallet = wl.wallet
        if key is not None:
            while True:
                try:
                    npass = getpass.getpass(prompt='Enter current password to encrypt ad-hoc key:')
                    crypt = AESCrypt(npass)
                    try:
                        decrypted = crypt.decrypt(wl.wallet.seed)
                        wallet = wl.wallet
                        wallet.seed = decrypted
                        password=npass
                    except DecryptionError:
                        print("**Invalid password**")
                except KeyboardInterrupt:
                    break
                    exit(0)

    if key is None:
        if count == 1:
            a = await wallet.account_create()
            print(f"account: {a}")
        else:
            async with in_transaction() as conn:
                ass = await wallet.accounts_create(count=count)
                for a in ass:
                    print(f"account: {a}")
    else:
        a = await wallet.adhoc_account_create(key, password=password)
        print(f"account: {a}")

async def wallet_destroy(wallet_id: str):
    # Retrieve wallet
    try:
        wallet = await Wallet.get_wallet(wallet_id)
    except WalletNotFound:
        print(f"No wallet found with ID: {wallet_id}")
        exit(1)
    except WalletLocked as wl:
        wallet = wl.wallet

    await wallet.delete()
    print("Wallet destroyed")

async def wallet_representative_get(wallet_id: str):
    # Retrieve wallet
    try:
        wallet = await Wallet.get_wallet(wallet_id)
    except WalletNotFound:
        print(f"No wallet found with ID: {wallet_id}")
        exit(1)
    except WalletLocked as wl:
        wallet = wl.wallet

    if wallet.representative is None:
        print("Representative not set")
    else:
        print(f"Wallet representative: {wallet.representative}")

async def wallet_representative_set(wallet_id: str, rep: str, update_existing: bool = False):
    # Retrieve wallet
    # Retrieve wallet
    crypt = None
    password=None
    if not Validators.is_valid_address(rep):
        print("Invalid representative")
        exit(1)
    try:
        wallet = await Wallet.get_wallet(wallet_id)
    except WalletNotFound:
        print(f"No wallet found with ID: {wallet_id}")
        exit(1)
    except WalletLocked as wl:
        wallet = wl.wallet
        if update_existing:
            while True:
                try:
                    npass = getpass.getpass(prompt='Enter current password to decrypt wallet:')
                    crypt = AESCrypt(npass)
                    try:
                        decrypted = crypt.decrypt(wl.wallet.seed)
                        wallet = wl.wallet
                        wallet.seed = decrypted
                        password=npass
                    except DecryptionError:
                        print("**Invalid password**")
                except KeyboardInterrupt:
                    break
                    exit(0)

    wallet.representative = rep
    await wallet.save(update_fields=['representative'])
    await wallet.bulk_representative_update(rep)
    print(f"Representative changed")

def main():
    loop = asyncio.new_event_loop()
    try:
        loop.run_until_complete(DBConfig().init_db())
        if options.command == 'wallet_list':
            loop.run_until_complete(wallet_list())
        elif options.command == 'wallet_create':
            if options.seed is not None:
                if not Validators.is_valid_block_hash(options.seed):
                    print("Invalid seed specified")
                    exit(1)
            loop.run_until_complete(wallet_create(options.seed))
        elif options.command == 'wallet_change_seed':
            if options.seed is not None:
                if not Validators.is_valid_block_hash(options.seed):
                    print("Invalid seed specified")
                    exit(1)
            else:
                while True:
                    try:
                        options.seed = getpass.getpass(prompt='Enter new wallet seed:')
                        if Validators.is_valid_block_hash(options.seed):
                            break
                        print("**Invalid seed**, should be a 64-character hex string")
                    except KeyboardInterrupt:
                        break
                        exit(0)
            password = ''
            if options.encrypt:
                while True:
                    try:
                        password = getpass.getpass(prompt='Enter password to encrypt wallet:')
                        if password.strip() == '':
                            print("**Bad password** - cannot be blanke")
                        break
                    except KeyboardInterrupt:
                        break
                        exit(0)
            loop.run_until_complete(wallet_change_seed(options.wallet, options.seed, password))
        elif options.command == 'wallet_view_seed':
            loop.run_until_complete(wallet_view_seed(options.wallet, options.password, options.all_keys))
        elif options.command == 'account_create':
            if options.key is not None:
                if not Validators.is_valid_block_hash(options.key):
                    print("Invalid Private Key")
                    exit(0)
            elif options.key is not None and options.count is not None:
                print("You can only specify one: --key or --count")
                print("--count can only be used for deterministic accounts")
            elif options.count is not None:
                if options.count < 1:
                    print("Count needs to be at least 1...")
            loop.run_until_complete(account_create(options.wallet, options.key, options.count))
        elif options.command == 'wallet_destroy':
            loop.run_until_complete(wallet_destroy(options.wallet))
        elif options.command == 'wallet_representative_get':
            loop.run_until_complete(wallet_representative_get(options.wallet))
        elif options.command == 'wallet_representative_set':
            loop.run_until_complete(wallet_representative_set(options.wallet, options.representatives, update_existing=options.update_existing))
        else:
            parser.print_help()
    except Exception as e:
        print(str(e))
        raise e
    finally:
        loop.run_until_complete(Tortoise.close_connections())
        loop.close()

if __name__ == "__main__":
    main()


# A simple bench mark for pippin
# 1) Create new wallet
# 2) Create 21 accounts
# 3) Load 1 account with 20 banano
# 4) Send 1 banano to 20 accounts
# 5) Consolidate it back into 1 account
# 6) Return to sender
import argparse
import asyncio
import datetime

from aiohttp import web, TCPConnector, ClientSession, AsyncResolver

parser = argparse.ArgumentParser(description="Pippin Benchmark")
parser.add_argument('--node-url', type=str, help='Node or Pippin RPC location', required=True)
options = parser.parse_args()

loop = asyncio.new_event_loop()

class RPCClient(object):
    _instance = None

    def __init__(self):
        raise RuntimeError('Call instance() instead')

    @classmethod
    def instance(cls) -> 'RPCClient':
        if cls._instance is None:
            cls._instance = cls.__new__(cls)
            cls.node_url = options.node_url
            cls.connector = TCPConnector(family=0 ,resolver=AsyncResolver())
            cls.session = ClientSession(connector=cls.connector)
        return cls._instance


    @classmethod
    async def close(cls):
        if hasattr(cls, 'session') and cls.session is not None:
            await cls.session.close()
        if cls._instance is not None:
            cls._instance = None

    async def make_request(self, req_json: dict):
        async with self.session.post(self.node_url ,json=req_json, timeout=300) as resp:
            return await resp.json()

async def do_send(rpc: RPCClient, source, destination, wallet):
    result = await rpc.make_request({
                'action': 'send',
                'source': source,
                'destination': destination,
                'amount': '100000000000000000000000000000',
                'wallet': wallet
            })
    return (result['block'], destination)

async def run_test():
    rpc = RPCClient.instance()
    # Wallet create
    wallet = (await rpc.make_request({
        'action':'wallet_create'
    }))['wallet']
    # Accounts_create
    accounts = (await rpc.make_request({
        'action':'accounts_create',
        'count': 21,
        'wallet': wallet
    }))['accounts']
    # Wait for user to send 50 BANANO to first account
    confirmed = False
    while not confirmed:
        try:
            print(f"Send 20 BANANO to {accounts[0]}")
            g = input("Type 'done' after it's been sent:")
            if g != 'done':
                print(f"Please type 'done' when you've sent it")
            else:
                confirmed = True
        except KeyboardInterrupt:
            break
    start_time = datetime.datetime.utcnow()
    pending_hash = (await rpc.make_request({
        'action':'pending',
        'account':accounts[0],
        'count':1,
        'include_active': True
    }))['blocks'][0]
    blocks_info = await rpc.make_request({
        'action': 'block_info',
        'json_block': True,
        'hash': pending_hash
    })
    sender = blocks_info['contents']['account']
    # Receive block
    print("receiving")
    received = (await rpc.make_request({
        'action':'receive',
        'wallet': wallet,
        'block': pending_hash,
        'account': accounts[0]
    }))['block']
    send_hashes = {}
    # Send to 20 accounts
    print("Distributing to accounts")
    tasks = []
    for a in accounts:
        if a == accounts[0]:
            continue
        tasks.append(do_send(rpc, accounts[0], a, wallet))
    ret = await asyncio.gather(*tasks)
    for r in ret:
        send_hashes[r[1]] = r[0]
    print("Receiving on all accounts")
    # Receive all blocks and send them back
    tasks = []
    for a, h in send_hashes.items():
        tasks.append(rpc.make_request({
            'action': 'receive',
            'wallet': wallet,
            'block': h,
            'account': a
        }))
    await asyncio.gather(*tasks)
    print(f"Conolidating back to {accounts[0]}")
    tasks = []
    for a, h in send_hashes.items():
        tasks.append(rpc.make_request({
            'action': 'send',
            'source': a,
            'destination': accounts[0],
            'amount': '100000000000000000000000000000',
            'wallet': wallet
        }))
    await asyncio.gather(*tasks)
    await asyncio.sleep(5)
    # Receive all pendings on account[0]
    pendings = (await rpc.make_request({
        'action':'pending',
        'account':accounts[0],
        'count':50,
        'include_active': True
    }))['blocks']
    tasks = []
    for p in pendings:
        tasks.append(rpc.make_request({
            'action': 'receive',
            'wallet': wallet,
            'block': p,
            'account': accounts[0]
        }))
    await asyncio.gather(*tasks)
    # Return to sender
    print(f"Returning to {sender}")
    send = (await rpc.make_request({
        'action': 'send',
        'source': accounts[0],
        'destination': sender,
        'amount': '2000000000000000000000000000000',
        'wallet': wallet
    }))['block']
    final_time = (datetime.datetime.utcnow() - start_time).total_seconds() - 5
    print(f"Finished in {final_time} seconds")

if __name__ == "__main__":
    loop.run_until_complete(run_test())
    loop.run_until_complete(RPCClient.close())
    loop.close()


import re

from bitstring import BitArray
from hashlib import blake2b

class Validators():
    @classmethod
    def is_valid_address(cls, input_text: str) -> bool:
        """Return True if address is valid, false otherwise"""
        if input_text is None:
            return False
        return cls.validate_checksum_xrb(input_text)

    @staticmethod
    def is_valid_block_hash(block_hash: str) -> bool:
        if block_hash is None or len(block_hash) != 64:
            return False
        try:
            int(block_hash, 16)
        except ValueError:
            return False
        return True

    @staticmethod
    def validate_checksum_xrb(address: str) -> bool:
        """Given an xrb/nano/ban address validate the checksum"""
        if (address[:5] == 'nano_' and len(address) == 65) or (address[:4] in ['ban_', 'xrb_']  and len(address) == 64):
            # Populate 32-char account index
            account_map = "13456789abcdefghijkmnopqrstuwxyz"
            account_lookup = {}
            for i in range(0, 32):
                account_lookup[account_map[i]] = BitArray(uint=i, length=5)

            # Extract key from address (everything after prefix)
            acrop_key = address[4:-8] if address[:5] != 'nano_' else address[5:-8]
            # Extract checksum from address
            acrop_check = address[-8:]

            # Convert base-32 (5-bit) values to byte string by appending each 5-bit value to the bitstring, essentially bitshifting << 5 and then adding the 5-bit value.
            number_l = BitArray()
            for x in range(0, len(acrop_key)):
                number_l.append(account_lookup[acrop_key[x]])
            number_l = number_l[4:]  # reduce from 260 to 256 bit (upper 4 bits are never used as account is a uint256)

            check_l = BitArray()
            for x in range(0, len(acrop_check)):
                check_l.append(account_lookup[acrop_check[x]])
            check_l.byteswap()  # reverse byte order to match hashing format

            # verify checksum
            h = blake2b(digest_size=5)
            h.update(number_l.bytes)
            return h.hexdigest() == check_l.hex
        return False


import nanolib
import pippin.config as config

class DifficultyModel(object):
  _instance = None

  def __init__(self):
    raise RuntimeError('Call instance() instead')

  @classmethod
  def instance(cls) -> 'DifficultyModel':
    if cls._instance is None:
      cls._instance = cls.__new__(cls)
      cls.receive_difficulty = 'fffffe0000000000' if config.Config.instance().banano else 'fffffe0000000000'
      cls.send_difficulty = 'fffffe0000000000' if config.Config.instance().banano else 'fffffff800000000'

    return cls._instance


from aiounittest import async_test
from pippin.util.nano_util import NanoUtil

import asyncio
import unittest

class TestNanoUtil(unittest.TestCase):
    @async_test
    async def test_work_generate(self):
        """Test work generation"""
        hash = "5E5B7C8F97BDA8B90FAA243050D99647F80C25EB4A07E7247114CBB129BDA188"
        difficulty = "ff00000000000000"
        result = await NanoUtil.instance().work_generate(hash, difficulty=difficulty)
        self.assertTrue(len(result) > 8)
        await NanoUtil.close()


from pippin.util.validators import Validators

import os
import string
import unittest

class TestValidators(unittest.TestCase):
    def test_valid_block_hash(self):
        # None type
        self.assertFalse(Validators.is_valid_block_hash(None))
        # Non-hex strings
        self.assertFalse(Validators.is_valid_block_hash('1234'))
        self.assertFalse(Validators.is_valid_block_hash('Appditto LLC'))
        self.assertFalse(Validators.is_valid_block_hash('ban_1bananobh5rat99qfgt1ptpieie5swmoth87thi74qgbfrij7dcgjiij94xr'))
        # Too short
        self.assertFalse(Validators.is_valid_block_hash('5E5B7C8F97BDA8B90FAA243050D99647F80C25EB4A07E7247114CBB129BDA18'))
        # Too long
        self.assertFalse(Validators.is_valid_block_hash('5E5B7C8F97BDA8B90FAA243050D99647F80C25EB4A07E7247114CBB129BDA1888'))
        # Non-hex character
        self.assertFalse(Validators.is_valid_block_hash('5E5B7C8F97BDA8B90FAA243050D99647F80C25EB4A07E7247114CBB129BDA18Z'))
        # Valid
        self.assertTrue(Validators.is_valid_block_hash('5E5B7C8F97BDA8B90FAA243050D99647F80C25EB4A07E7247114CBB129BDA188'))

    def test_valid_address(self):
        """Test address validation"""
        # Null should always be false
        self.assertFalse(Validators.is_valid_address(None))
        os.environ['BANANO'] = '1'
        # Valid
        self.assertTrue(Validators.is_valid_address('ban_1bananobh5rat99qfgt1ptpieie5swmoth87thi74qgbfrij7dcgjiij94xr'))
        # Bad checksum
        self.assertFalse(Validators.is_valid_address('ban_1bananobh5rat99qfgt1ptpieie5swmoth87thi74qgbfrij7dcgjiij94xa'))
        # Bad length
        self.assertFalse(Validators.is_valid_address('ban_1bananobh5rat99qfgt1ptpieie5swmoth87thi74qgbfrij7dcgjiij94x'))
        del os.environ['BANANO']
        # Valid
        self.assertTrue(Validators.is_valid_address('nano_1bananobh5rat99qfgt1ptpieie5swmoth87thi74qgbfrij7dcgjiij94xr'))
        # Bad checksum
        self.assertFalse(Validators.is_valid_address('nano_1bananobh5rat99qfgt1ptpieie5swmoth87thi74qgbfrij7dcgjiij94xa'))
        # Bad length
        self.assertFalse(Validators.is_valid_address('nano_1bananobh5rat99qfgt1ptpieie5swmoth87thi74qgbfrij7dcgjiij94x'))
        # Valid
        self.assertTrue(Validators.is_valid_address('xrb_1bananobh5rat99qfgt1ptpieie5swmoth87thi74qgbfrij7dcgjiij94xr'))
        # Bad checksum
        self.assertFalse(Validators.is_valid_address('xrb_1bananobh5rat99qfgt1ptpieie5swmoth87thi74qgbfrij7dcgjiij94xa'))
        # Bad length
        self.assertFalse(Validators.is_valid_address('xrb_1bananobh5rat99qfgt1ptpieie5swmoth87thi74qgbfrij7dcgjiij94x'))



import pathlib
from dotenv import load_dotenv
load_dotenv()
from pippin.util.utils import Utils
load_dotenv(dotenv_path=Utils.get_project_root().joinpath(pathlib.PurePath('.env')))

from aiohttp import log
from pippin.version import __version__
from pippin.util.utils import Utils
from pippin.util.validators import Validators

import secrets
import yaml

DEFAULT_BANANO_REPS = [
    'ban_1ka1ium4pfue3uxtntqsrib8mumxgazsjf58gidh1xeo5te3whsq8z476goo',
    'ban_1cake36ua5aqcq1c5i3dg7k8xtosw7r9r7qbbf5j15sk75csp9okesz87nfn',
    'ban_1fomoz167m7o38gw4rzt7hz67oq6itejpt4yocrfywujbpatd711cjew8gjj'
]

DEFAULT_NANO_REPS = [
    'nano_1x7biz69cem95oo7gxkrw6kzhfywq4x5dupw4z1bdzkb74dk9kpxwzjbdhhs',
    'nano_1thingspmippfngcrtk1ofd3uwftffnu4qu9xkauo9zkiuep6iknzci3jxa6',
    'nano_1natrium1o3z5519ifou7xii8crpxpk8y65qmkih8e8bpsjri651oza8imdd',
    'nano_3o7uzba8b9e1wqu5ziwpruteyrs3scyqr761x7ke6w1xctohxfh5du75qgaj'
]

class Config(object):
    _instance = None

    def __init__(self):
        raise RuntimeError('Call instance() instead')

    @classmethod
    def instance(cls) -> 'Config':
        if cls._instance is None:
            cls._instance = cls.__new__(cls)
            try:
                with open(f"{Utils.get_project_root().joinpath(pathlib.PurePath('config.yaml'))}", "r") as in_yaml:
                    cls.yaml = list(yaml.load_all(in_yaml, Loader=yaml.FullLoader))[0]
            except FileNotFoundError:
                cls.yaml = None
            # Parse options
            cls.banano = cls.get_yaml_property('wallet', 'banano', False)
            cls.log_file = cls.get_yaml_property('server', 'log_file', default='/tmp/pippin_wallet.log')
            cls.debug = cls.get_yaml_property('server', 'debug', default=False)
            cls.stdout = cls.get_yaml_property('server', 'log_to_stdout', default=False)
            cls.node_url = cls.get_yaml_property('server', 'node_rpc_url', default='http://[::1]:7072' if cls.banano else 'http://[::1]:7076')
            cls.node_ws_url = cls.get_yaml_property('server', 'node_ws_url', None)
            cls.port = cls.get_yaml_property('server', 'port', default=11338)
            cls.host = cls.get_yaml_property('server', 'host', default='127.0.0.1')
            cls.work_peers = cls.get_yaml_property('wallet', 'work_peers', [])
            cls.node_work_generate = cls.get_yaml_property('wallet', 'node_work_generate', False)
            cls.receive_minimum = cls.get_yaml_property('wallet', 'receive_minimum', 1000000000000000000000000000 if cls.banano else 1000000000000000000000000)
            cls.auto_receive_on_send = cls.get_yaml_property('wallet', 'auto_receive_on_send', True)
            cls.max_work_processes = cls.get_yaml_property('wallet', 'max_work_processes', 1)
            cls.max_sign_threads = cls.get_yaml_property('wallet', 'max_sign_threads', 1)
            if not cls.banano:
                cls.preconfigured_reps = cls.get_yaml_property('wallet', 'preconfigured_representatives_nano', default=None)
            else:
                cls.preconfigured_reps = cls.get_yaml_property('wallet', 'preconfigured_representatives_banano', default=None)
            # Enforce that all reps are valid
            if cls.preconfigured_reps is not None:
                cls.preconfigured_reps = set(cls.preconfigured_reps)
                for r in cls.preconfigured_reps:
                    if not Validators.is_valid_address(r):
                        log.server_logger.warn(f"{r} is not a valid representative!")
                        cls.preconfigured_reps.remove(r)
                if len(cls.preconfigured_reps) == 0:
                    cls.preconfigured_reps = None
            # Go to default if None
            if cls.preconfigured_reps is None:
                cls.preconfigured_reps = DEFAULT_BANANO_REPS if cls.banano else DEFAULT_NANO_REPS

        return cls._instance

    @classmethod
    def get_yaml_property(cls, category: str, subcategory: str, default):
        """Get a property from yaml config"""
        if cls.yaml is None:
            return default
        elif category in cls.yaml and cls.yaml[category] is not None and subcategory in cls.yaml[category]:
            return cls.yaml[category][subcategory]
        return default

    def get_random_rep(self) -> str:
        """Returns a random representative"""
        return secrets.choice(self.preconfigured_reps)



import datetime
import logging
from json import JSONDecodeError

import rapidjson as json
from aiohttp import log, web
from tortoise.transactions import in_transaction

import asyncio
import pippin.config as config
from pippin.db.redis import RedisDB
from pippin.db.models.account import Account
from pippin.db.models.adhoc_account import AdHocAccount
from pippin.db.models.wallet import (AccountAlreadyExists, Wallet, WalletLocked,
                              WalletNotFound)
from pippin.model.difficulty import DifficultyModel
from pippin.network.rpc_client import AccountNotFound, BlockNotFound, RPCClient
from pippin.network.nano_websocket import WebsocketClient
from pippin.network.work_client import WorkClient
from pippin.util.crypt import DecryptionError
from pippin.util.random import RandomUtil
from pippin.util.validators import Validators
from pippin.util.wallet import (InsufficientBalance, ProcessFailed, WalletUtil,
                         WorkFailed)

class PippinServer(object):
    """API for wallet requests"""
    def __init__(self, host: str, port: int):
        self.app = web.Application(middlewares=[web.normalize_path_middleware()])
        self.app.add_routes([
            web.post('/', self.gateway)
        ])
        self.host = host
        self.port = port
        self.websocket = None
        if config.Config.instance().node_ws_url is not None:
            self.websocket = WebsocketClient(config.Config.instance().node_ws_url, self.block_arrival_handler)

    async def stop(self):
        await self.app.shutdown()
        if self.websocket:
            await self.websocket.close()

    def json_response(self, data: dict):
        """Wrapper for json responses using custom json parser"""
        return web.json_response(
            data=data,
            dumps=json.dumps
        )

    def generic_error(self):
        """The node returns this generic error when the request is bad"""
        return self.json_response(
            data={
                'error':"Unable to parse json"
            }
        )

    async def gateway(self, request: web.Request):
        """Gateway route to mimic nano's API of specifying action in a string"""
        try:
            request_json = await request.json(loads=json.loads)    
        except json.JSONDecodeError:
            return self.generic_error()
        if 'action' in request_json:
            # Sanitize action
            request_json['action'] = request_json['action'].lower().strip()

            # Handle wallet RPCs
            if request_json['action'] == 'wallet_create':
                return await self.wallet_create(request, request_json)
            elif request_json['action'] == 'account_create':
                return await self.account_create(request, request_json)
            elif request_json['action'] == 'accounts_create':
                return await self.accounts_create(request, request_json)
            elif request_json['action'] == 'account_list':
                return await self.account_list(request, request_json)
            elif request_json['action'] == 'receive':
                return await self.receive(request, request_json)
            elif request_json['action'] == 'send':
                return await self.send(request, request_json)
            elif request_json['action'] == 'account_representative_set':
                return await self.account_representative_set(request, request_json)
            elif request_json['action'] == 'password_change':
                return await self.password_change(request, request_json)
            elif request_json['action'] == 'password_enter':
                return await self.password_enter(request, request_json)
            elif request_json['action'] == 'password_valid':
                return await self.password_valid(request, request_json)
            elif request_json['action'] == 'wallet_representative_set':
                return await self.wallet_representative_set(request, request_json)
            elif request_json['action'] == 'wallet_add':
                return await self.wallet_add(request, request_json)
            elif request_json['action'] == 'wallet_lock':
                return await self.wallet_lock(request, request_json)
            elif request_json['action'] == 'wallet_locked':
                return await self.wallet_locked(request, request_json)
            elif request_json['action'] == 'wallet_balances':
                return await self.wallet_balances(request, request_json)
            elif request_json['action'] == 'wallet_frontiers':
                return await self.wallet_frontiers(request, request_json)
            elif request_json['action'] == 'wallet_pending':
                return await self.wallet_pending(request, request_json)
            elif request_json['action'] == 'wallet_destroy':
                return await self.wallet_destroy(request, request_json)
            elif request_json['action'] == 'wallet_change_seed':
                return await self.wallet_change_seed(request, request_json)
            elif request_json['action'] == 'wallet_contains':
                return await self.wallet_contains(request, request_json)
            elif request_json['action'] == 'wallet_representative':
                return await self.wallet_representative(request, request_json)
            elif request_json['action'] == 'wallet_info':
                return await self.wallet_info(request, request_json)
            elif request_json['action'] == 'receive_all':
                return await self.receive_all(request, request_json)
            elif request_json['action'] == 'work_generate':
                return await self.work_generate(request, request_json)
            elif request_json['action'] in ['account_move', 'account_remove', 'receive_minimum', 'receive_minimum_set', 'search_pending', 'search_pending_all', 'wallet_add_watch', 'wallet_export', 'wallet_history', 'wallet_ledger', 'wallet_republish', 'wallet_work_get', 'work_get', 'work_set']:
                # Prevent unimplemented wallet RPCs from going to the node directly
                return self.json_response(
                    data = {
                        'error': 'not_implemented'
                    }
                )

            # Proxy other requests to the node
            resp_json = await RPCClient.instance().make_request(request_json)
            return self.json_response(
                data = resp_json
            )

        return self.generic_error()

    async def wallet_create(self, request: web.Request, request_json: dict):
        """Route for creating new wallet"""
        if 'seed' in request_json:
            if not Validators.is_valid_block_hash(request_json['seed']):
                return self.json_response(
                    data = {'error': 'Invalid seed'}
                )
            new_seed = request_json['seed']
        else:
            new_seed = RandomUtil.generate_seed()
        async with in_transaction() as conn:
            wallet = Wallet(
                seed=new_seed
            )
            await wallet.save(using_db=conn)
            await wallet.account_create(using_db=conn)
        return self.json_response(
            data = {
                'wallet': str(wallet.id)
            }
        )

    async def account_create(self, request: web.Request, request_json: dict):
        """Route for creating new wallet"""
        if 'wallet' not in request_json:
            return self.generic_error()

        try:
            wallet = await Wallet.get_wallet(request_json['wallet'])
        except WalletNotFound:
            return self.json_response(
                data={
                    'error': 'wallet not found'
                }
            )
        except WalletLocked:
            return self.json_response(
                data={
                    'error': 'wallet locked'
                }
            )

        # Create account
        async with in_transaction() as conn:
            account = await wallet.account_create(using_db=conn)
        return self.json_response(
            data = {
                'account': account
            }
        )

    async def accounts_create(self, request: web.Request, request_json: dict):
        """Route for creating new wallet"""
        if 'wallet' not in request_json or 'count' not in request_json or not isinstance(request_json['count'], int):
            return self.generic_error()

        try:
            wallet = await Wallet.get_wallet(request_json['wallet'])
        except WalletNotFound:
            return self.json_response(
                data={
                    'error': 'wallet not found'
                }
            )
        except WalletLocked:
            return self.json_response(
                data={
                    'error': 'wallet locked'
                }
            )

        # Create account
        async with in_transaction() as conn:
            accounts = await wallet.accounts_create(count=request_json['count'], using_db=conn)
        return self.json_response(
            data = {
                'accounts': accounts
            }
        )

    async def account_list(self, request: web.Request, request_json: dict):
        """Route for creating new wallet"""
        if 'wallet' not in request_json:
            return self.generic_error()
        elif 'count' in request_json and isinstance(request_json['count'], int):
            count = request_json['count']
        else:
            count = 1000

        try:
            wallet = await Wallet.get_wallet(request_json['wallet'])
        except WalletNotFound:
            return self.json_response(
                data={
                    'error': 'wallet not found'
                }
            )
        except WalletLocked:
            return self.json_response(
                data={
                    'error': 'wallet locked'
                }
            )

        return self.json_response(
            data = {'accounts': [a.address for a in await wallet.accounts.all().limit(count)]}
        )

    async def receive(self, request: web.Request, request_json: dict):
        """RPC receive"""
        if 'wallet' not in request_json or 'account' not in request_json or 'block' not in request_json:
            return self.generic_error()
        elif not Validators.is_valid_address(request_json['account']):
            return self.json_response(
                data={'error': 'Invalid address'}
            )
        elif not Validators.is_valid_block_hash(request_json['block']):
            return self.json_response(
                data={'error': 'Invalid block'}
            )

        work = request_json['work'] if 'work' in request_json else None

        # Retrieve wallet
        try:
            wallet = await Wallet.get_wallet(request_json['wallet'])
        except WalletNotFound:
            return self.json_response(
                data={
                    'error': 'wallet not found'
                }
            )
        except WalletLocked:
            return self.json_response(
                data={
                    'error': 'wallet locked'
                }
            )

        # Retrieve account on wallet
        account = await wallet.get_account(request_json['account'])
        if account is None:
            return self.json_response(
                data={'error': 'Account not found'}
            )

        # Try to receive block
        wallet = WalletUtil(account, wallet)
        try:
            response = await wallet.receive(request_json['block'], work=work)
        except BlockNotFound:
            return self.json_response(
                data={'error': 'Block not found'}
            )
        except WorkFailed:
            return self.json_response(
                data={'error': 'Failed to generate work'}
            )
        except ProcessFailed:
            return self.json_response(
                data={'error': 'RPC Process failed'}
            )

        if response is None:
            return self.json_response(
                data={'error': 'Unable to receive block'}
            )

        return self.json_response(
            data=response
        )

    async def send(self, request: web.Request, request_json: dict):
        """RPC send"""
        if 'wallet' not in request_json or 'source' not in request_json or 'destination' not in request_json or 'amount' not in request_json:
            return self.generic_error()
        elif not Validators.is_valid_address(request_json['source']):
            return self.json_response(
                data={'error': 'Invalid source'}
            )
        elif not Validators.is_valid_address(request_json['destination']):
            return self.json_response(
                data={'error': 'Invalid destination'}
            )

        id = request_json['id'] if 'id' in request_json else None
        work = request_json['work'] if 'work' in request_json else None

        # Retrieve wallet
        try:
            wallet = await Wallet.get_wallet(request_json['wallet'])
        except WalletNotFound:
            return self.json_response(
                data={
                    'error': 'wallet not found'
                }
            )
        except WalletLocked:
            return self.json_response(
                data={
                    'error': 'wallet locked'
                }
            )

        # Retrieve account on wallet
        account = await wallet.get_account(request_json['source'])
        if account is None:
            return self.json_response(
                data={'error': 'Account not found'}
            )

        # Try to create and publish send block
        wallet = WalletUtil(account, wallet)
        try:
            resp = await wallet.send(int(request_json['amount']), request_json['destination'], id=id, work=work)
        except AccountNotFound:
            return self.json_response(
                data={'error': 'Account not found'}
            )
        except BlockNotFound:
            return self.json_response(
                data={'error': 'Block not found'}
            )
        except WorkFailed:
            return self.json_response(
                data={'error': 'Failed to generate work'}
            )
        except ProcessFailed:
            return self.json_response(
                data={'error': 'RPC Process failed'}
            )
        except InsufficientBalance:
            return self.json_response(
                data={'error': 'insufficient balance'}
            )

        if resp is None:
            return self.json_response(
                data={'error': 'Unable to create send block'}
            )

        return self.json_response(
            data=resp
        )

    async def account_representative_set(self, request: web.Request, request_json: dict):
        """RPC account_representative_set"""
        if 'wallet' not in request_json or 'account' not in request_json or 'representative' not in request_json:
            return self.generic_error()
        elif not Validators.is_valid_address(request_json['account']):
            return self.json_response(
                data={'error': 'Invalid account'}
            )
        elif not Validators.is_valid_address(request_json['representative']):
            return self.json_response(
                data={'error': 'Invalid representative'}
            )

        work = request_json['work'] if 'work' in request_json else None

        # Retrieve wallet
        try:
            wallet = await Wallet.get_wallet(request_json['wallet'])
        except WalletNotFound:
            return self.json_response(
                data={
                    'error': 'wallet not found'
                }
            )
        except WalletLocked:
            return self.json_response(
                data={
                    'error': 'wallet locked'
                }
            )

        # Retrieve account on wallet
        account = await wallet.get_account(request_json['account'])
        if account is None:
            return self.json_response(
                data={'error': 'Account not found'}
            )

        # Try to create and publish CHANGE block
        wallet = WalletUtil(account, wallet)
        try:
            resp = await wallet.representative_set(request_json['representative'], work=work)
        except AccountNotFound:
            return self.json_response(
                data={'error': 'Account not found'}
            )
        except WorkFailed:
            return self.json_response(
                data={'error': 'Failed to generate work'}
            )
        except ProcessFailed:
            return self.json_response(
                data={'error': 'RPC Process failed'}
            )

        if resp is None:
            return self.json_response(
                data={'error': 'Unable to create change block'}
            )

        return self.json_response(
            data=resp
        )

    async def password_change(self, request: web.Request, request_json: dict):
        """RPC password_change"""
        if 'wallet' not in request_json or 'password' not in request_json:
            return self.generic_error()

        # Retrieve wallet
        try:
            wallet = await Wallet.get_wallet(request_json['wallet'])
        except WalletNotFound:
            return self.json_response(
                data={
                    'error': 'wallet not found'
                }
            )
        except WalletLocked:
            return self.json_response(
                data={
                    'error': 'wallet locked'
                }
            )

        # Encrypt
        await wallet.encrypt_wallet(request_json['password'])

        return self.json_response(
            data={'changed': '1'}
        )

    async def password_enter(self, request: web.Request, request_json: dict):
        """RPC password_enter"""
        if 'wallet' not in request_json or 'password' not in request_json:
            return self.generic_error()

        # Retrieve wallet
        try:
            wallet = await Wallet.get_wallet(request_json['wallet'])
            return self.json_response(
                data={
                    'error': 'wallet not locked'
                }
            )
        except WalletNotFound:
            return self.json_response(
                data={
                    'error': 'wallet not found'
                }
            )
        except WalletLocked as w:
            wallet = w.wallet

        try:
            await wallet.unlock_wallet(request_json['password'])
        except DecryptionError:
            return self.json_response(
                data={'valid': '0'}
            )

        return self.json_response(
            data={'valid': '1'}
        )

    async def password_valid(self, request: web.Request, request_json: dict):
        """RPC password_valid"""
        if 'wallet' not in request_json:
            return self.generic_error()

        # Retrieve wallet
        try:
            wallet = await Wallet.get_wallet(request_json['wallet'])
            if not wallet.encrypted:
                return self.json_response(
                    data={
                        'error': 'wallet not locked'
                    }
                )
        except WalletNotFound:
            return self.json_response(
                data={
                    'error': 'wallet not found'
                }
            )
        except WalletLocked:
            return self.json_response(
                data={'valid': '0'}
            )

        return self.json_response(
            data={'valid': '1'}
        )

    async def wallet_representative_set(self, request: web.Request, request_json: dict):
        """RPC wallet_representative_set"""
        if 'wallet' not in request_json or 'representative' not in request_json or ('update_existing_accounts' in request_json and not isinstance(request_json['update_existing_accounts'], bool)):
            return self.generic_error()
        elif not Validators.is_valid_address(request_json['representative']):
            return self.json_response(
                data={'error': 'Invalid address'}
            )

        update_existing = False
        if 'update_existing_accounts' in request_json:
            update_existing = request_json['update_existing_accounts']

        # Retrieve wallet
        try:
            wallet = await Wallet.get_wallet(request_json['wallet'])
        except WalletNotFound:
            return self.json_response(
                data={
                    'error': 'wallet not found'
                }
            )
        except WalletLocked:
            return self.json_response(
                data={
                    'error': 'wallet locked'
                }
            )

        wallet.representative = request_json['representative']
        await wallet.save(update_fields=['representative'])

        if update_existing:
            await wallet.bulk_representative_update(request_json['representative'])

        return self.json_response(
            data={'set': '1'}
        )

    async def wallet_add(self, request: web.Request, request_json: dict):
        """RPC wallet_add"""
        if 'wallet' not in request_json or 'key' not in request_json:
            return self.generic_error()
        elif not Validators.is_valid_block_hash(request_json['key']):
            return self.json_response(
                data={'error': 'Invalid key'}
            )

        # Retrieve wallet
        try:
            wallet = await Wallet.get_wallet(request_json['wallet'])
        except WalletNotFound:
            return self.json_response(
                data={
                    'error': 'wallet not found'
                }
            )
        except WalletLocked:
            return self.json_response(
                data={
                    'error': 'wallet locked'
                }
            )

        # Add account
        try:
            address = await wallet.adhoc_account_create(request_json['key'])
        except AccountAlreadyExists:
            return self.json_response(
                data={
                    'error': 'account already exists'
                }
            )

        return self.json_response(
            data={'account':address}
        )

    async def wallet_lock(self, request: web.Request, request_json: dict):
        """RPC wallet_lock"""
        if 'wallet' not in request_json:
            return self.generic_error()

        # Retrieve wallet
        try:
            wallet = await Wallet.get_wallet(request_json['wallet'])
        except WalletNotFound:
            return self.json_response(
                data={
                    'error': 'wallet not found'
                }
            )
        except WalletLocked as we:
            wallet = we.wallet

        await wallet.lock_wallet()

        return self.json_response(
            data={'locked':'1'}
        )

    async def wallet_locked(self, request: web.Request, request_json: dict):
        """RPC wallet_locked"""
        if 'wallet' not in request_json:
            return self.generic_error()

        # Retrieve wallet
        try:
            wallet = await Wallet.get_wallet(request_json['wallet'])
        except WalletNotFound:
            return self.json_response(
                data={
                    'error': 'wallet not found'
                }
            )
        except WalletLocked:
            return self.json_response(
                data={'locked': '1'}
            )

        return self.json_response(
            data={'locked':'0'}
        )

    async def wallet_balances(self, request: web.Request, request_json: dict):
        """RPC wallet_balances"""
        if 'wallet' not in request_json:
            return self.generic_error()

        # Retrieve wallet
        try:
            wallet = await Wallet.get_wallet(request_json['wallet'])
        except WalletNotFound:
            return self.json_response(
                data={
                    'error': 'wallet not found'
                }
            )
        except WalletLocked:
            return self.json_response(
                data={
                    'error': 'wallet locked'
                }
            )

        return self.json_response(
            data=await RPCClient.instance().accounts_balances([a.address for a in await wallet.get_all_accounts()])
        )

    async def wallet_pending(self, request: web.Request, request_json: dict):
        """RPC wallet_pending"""
        if 'wallet' not in request_json:
            return self.generic_error()

        # Retrieve wallet
        try:
            wallet = await Wallet.get_wallet(request_json['wallet'])
        except WalletNotFound:
            return self.json_response(
                data={
                    'error': 'wallet not found'
                }
            )
        except WalletLocked:
            return self.json_response(
                data={
                    'error': 'wallet locked'
                }
            )

        return self.json_response(
            data=await RPCClient.instance().accounts_pending([a.address for a in await wallet.get_all_accounts()])
        )

    async def wallet_frontiers(self, request: web.Request, request_json: dict):
        """RPC wallet_frontiers"""
        if 'wallet' not in request_json:
            return self.generic_error()

        # Retrieve wallet
        try:
            wallet = await Wallet.get_wallet(request_json['wallet'])
        except WalletNotFound:
            return self.json_response(
                data={
                    'error': 'wallet not found'
                }
            )
        except WalletLocked:
            return self.json_response(
                data={
                    'error': 'wallet locked'
                }
            )

        return self.json_response(
            data=await RPCClient.instance().accounts_frontiers([a.address for a in await wallet.get_all_accounts()])
        )

    async def wallet_destroy(self, request: web.Request, request_json: dict):
        """RPC wallet_destroy"""
        if 'wallet' not in request_json:
            return self.generic_error()

        # Retrieve wallet
        try:
            wallet = await Wallet.get_wallet(request_json['wallet'])
        except WalletNotFound:
            return self.json_response(
                data={
                    'error': 'wallet not found'
                }
            )
        except WalletLocked:
            return self.json_response(
                data={
                    'error': 'wallet locked'
                }
            )
        await wallet.delete()
        return self.json_response(
            data={'destroyed': '1'}
        )

    async def wallet_change_seed(self, request: web.Request, request_json: dict):
        """RPC wallet_change_seed"""
        if 'wallet' not in request_json or 'seed' not in request_json:
            return self.generic_error()
        elif not Validators.is_valid_block_hash(request_json['seed']):
            return self.json_response(
                data={'error': 'Invalid seed'}
            )

        # Retrieve wallet
        try:
            wallet = await Wallet.get_wallet(request_json['wallet'])
        except WalletNotFound:
            return self.json_response(
                data={
                    'error': 'wallet not found'
                }
            )
        except WalletLocked:
            return self.json_response(
                data={
                    'error': 'wallet locked'
                }
            )

        # Reset password
        if wallet.encrypted:
            await wallet.encrypt_wallet('')

        # Change key
        await wallet.change_seed(request_json['seed'])

        # Get newest account
        newest = await wallet.get_newest_account()

        return self.json_response(
            data={
                "success": "",
                "last_restored_account": newest.address,
                "restored_count": newest.account_index + 1
            }
        )

    async def wallet_contains(self, request: web.Request, request_json: dict):
        """RPC wallet_contains"""
        if 'wallet' not in request_json or 'account' not in request_json:
            return self.generic_error()
        elif not Validators.is_valid_address(request_json['account']):
            return self.json_response(
                data={'error': 'Invalid account'}
            )

        # Retrieve wallet
        try:
            wallet = await Wallet.get_wallet(request_json['wallet'])
        except WalletNotFound:
            return self.json_response(
                data={
                    'error': 'wallet not found'
                }
            )
        except WalletLocked:
            return self.json_response(
                data={
                    'error': 'wallet locked'
                }
            )

        exists = (await wallet.get_account(request_json['account'])) is not None

        return self.json_response(
            data={'exists': '1' if exists else '0'}
        )

    async def wallet_representative(self, request: web.Request, request_json: dict):
        """RPC wallet_representative"""
        if 'wallet' not in request_json:
            return self.generic_error()

        # Retrieve wallet
        try:
            wallet = await Wallet.get_wallet(request_json['wallet'])
        except WalletNotFound:
            return self.json_response(
                data={
                    'error': 'wallet not found'
                }
            )
        except WalletLocked:
            return self.json_response(
                data={
                    'error': 'wallet locked'
                }
            )

        if wallet.representative is None:
            wallet.representative = config.Config.instance().get_random_rep()

        return self.json_response(
            data={'representative': wallet.representative}
        )

    async def wallet_info(self, request: web.Request, request_json: dict):
        """RPC wallet_info"""
        if 'wallet' not in request_json:
            return self.generic_error()

        # Retrieve wallet
        try:
            wallet = await Wallet.get_wallet(request_json['wallet'])
        except WalletNotFound:
            return self.json_response(
                data={
                    'error': 'wallet not found'
                }
            )
        except WalletLocked:
            return self.json_response(
                data={
                    'error': 'wallet locked'
                }
            )

        # Get balances
        balance = 0
        pending_bal = 0
        balance_json = await RPCClient.instance().accounts_balances([a.address for a in await wallet.get_all_accounts()])

        if 'balances' not in balance_json:
            return self.json_response(
                data={'error': 'failed to retrieve balances'}
            )
    
        for k, v in balance_json['balances'].items():
            balance += int(v['balance'])
            pending_bal += int(v['pending'])

        det_count = await wallet.accounts.all().count()
        adhoc_count = await wallet.accounts.all().count()
        det_index = (await wallet.get_newest_account()).account_index

        return self.json_response(
            data={
                'balance': balance,
                'pending': pending_bal,
                'accounts_count': det_count + adhoc_count,
                'adhoc_count': adhoc_count,
                'deterministic_count': det_count,
                'deterministic_index': det_index
            }
        )

    async def receive_all(self, request: web.Request, request_json: dict):
        """receive every pending block in wallet"""
        if 'wallet' not in request_json:
            return self.generic_error()

        # Retrieve wallet
        try:
            wallet = await Wallet.get_wallet(request_json['wallet'])
        except WalletNotFound:
            return self.json_response(
                data={
                    'error': 'wallet not found'
                }
            )
        except WalletLocked:
            return self.json_response(
                data={
                    'error': 'wallet locked'
                }
            )

        balance_json = await RPCClient.instance().accounts_balances([a.address for a in await wallet.get_all_accounts()])

        if 'balances' not in balance_json:
            return self.json_response(
                data={'error': 'failed to retrieve balances'}
            )
    
        received_count = 0
        for k, v in balance_json['balances'].items():
            if int(v['pending']) > 0:
                wallet_util = WalletUtil(await wallet.get_account(k), wallet)
                received_count += await wallet_util.receive_all()

        return self.json_response(
            data={
                'received': received_count
            }
        )

    async def work_generate(self, request: web.Request, request_json: dict):
        """Route for running work_generate"""
        if 'hash' in request_json:
            if not Validators.is_valid_block_hash(request_json['hash']):
                return self.json_response(
                    data = {'error': 'Invalid hash'}
                )
        else:
            return self.generic_error()

        difficulty = DifficultyModel.instance().send_difficulty
        if 'difficulty' in request_json:
            difficulty = request_json['difficulty']
        elif 'subtype' in request_json and request_json['subtype'] == 'receive':
            difficulty = DifficultyModel.instance().receive_difficulty

        # Generate work
        work = await WorkClient.instance().work_generate(request_json['hash'], difficulty)
        if work is None:
            return self.json_response(
                data = {
                    'error': 'Failed to generate work'
                }
            )
        return self.json_response(
            data = {
                'work': work
            }
        )

    async def block_arrival_handler(self, data: dict):
        """invoked when we receive a new block"""
        log.server_logger.debug("Received Callback")
        is_send = False
        if 'block' in data and 'subtype' in data['block'] and data['block']['subtype'] == 'send':
            is_send = True
        elif 'is_send' in data and (data['is_send'] == 'true' or data['is_send']):
            is_send = True
        if is_send:
            # Ignore receive_minimum
            if config.Config.instance().receive_minimum > int(data['amount']):
                return
            # Determine if the recipient is one of ours
            destination = data['block']['link_as_account']
            acct = await Account.filter(address=destination).prefetch_related('wallet').first()
            if acct is None:
                acct = await AdHocAccount.filter(address=destination).prefetch_related('wallet').first()
                if acct is None:
                    return
            log.server_logger.debug(f"Auto receiving {data['hash']} for {destination}")
            wu = WalletUtil(acct, acct.wallet)
            try:
                await wu.receive(data['hash'])
            except Exception:
                log.server_logger.debug(f"Failed to receive {data['hash']}")

    async def start(self):
        """Start the server"""
        runner = web.AppRunner(self.app, access_log = None if not config.Config.instance().debug else log.server_logger)
        await runner.setup()
        site = web.TCPSite(runner, self.host, self.port)
        tasks = []
        tasks.append(site.start())
        # Websocket
        if self.websocket:
            await self.websocket.setup()
            tasks.append(self.websocket.loop())
        await asyncio.wait(tasks)



import asyncio
import os

import aioredis
import aioredlock

class RedisDB(object):
    _instance = None

    def __init__(self):
        raise RuntimeError('Call instance() instead')

    @classmethod
    def instance(cls) -> 'RedisDB':
        if cls._instance is None:
            cls._instance = cls.__new__(cls)
            cls.redis = None
            cls.lock_manager = None
            cls.redis_host = os.getenv('REDIS_HOST', '127.0.0.1')
            cls.redis_port = int(os.getenv('REDIS_PORT', '6379'))
            cls.redis_db = int(os.getenv('REDIS_DB', 0))
            cls.redis_pw = os.getenv('REDIS_PW', None)
        return cls._instance

    @classmethod
    async def close(cls):
        if hasattr(cls, 'redis') and cls.redis is not None:
            cls.redis.close()
            await cls.redis.wait_closed()
        if cls._instance is not None:
            cls._instance = None

    @classmethod
    async def get_redis(cls) -> aioredis.Redis:
        if cls.redis is not None:
            return cls.redis
        cls.redis = await aioredis.create_redis_pool((cls.redis_host, cls.redis_port), db=cls.redis_db, password=cls.redis_pw, encoding='utf-8', minsize=1, maxsize=5)
        return cls.redis

    @classmethod
    async def get_lock_manager(cls) -> aioredlock.Aioredlock:
        if cls.lock_manager is not None:
            return cls.lock_manager
        cls.lock_manager = aioredlock.Aioredlock([await cls.get_redis()], lock_timeout=300, retry_count=3, retry_delay_min=30, retry_delay_max=90)
        return cls.lock_manager

    async def set(self, key: str, value: str, expires: int = 0):
        """Basic redis SET"""
        # Add a prefix
        key = f"pippin:{key}"
        redis = await self.get_redis()
        await redis.set(key, value, expire=expires)

    async def get(self, key: str):
        """Redis GET"""
        # Add a prefix
        key = f"pippin:{key}"
        redis = await self.get_redis()
        return await redis.get(key)

    async def delete(self, key: str):
        """Redis DELETE"""
        key = f"pippin:{key}"
        await self._delete(key)

    async def _delete(self, key: str):
        """Redis DELETE"""
        redis = await self.get_redis()
        await redis.delete(key)

    async def exists(self, key: str):
        """See if a key exists"""
        key = f"pippin:{key}"
        redis = await self.get_redis()
        return (await redis.get(key)) is not None



# coding: utf8
import re
import sys
import pippin.version

from setuptools import find_packages, setup

if sys.version_info < (3, 6):
    raise RuntimeError("Pippin requires Python >= 3.6")

def requirements() -> list:
    try:
        ret = open("requirements.txt", "rt").read().splitlines()
        if sys.platform not in ('win32', 'cygwin', 'cli'):
            ret.append('uvloop>=0.14.0')
        return ret
    except FileNotFoundError:
        ret = [
            'tortoise-orm>=0.15.24,<0.16',
            'aiosqlite>=0.10.0',
            'asyncpg>=0.20.0',
            'pypika<0.37.0',
            'aiomysql>=0.0.20',
            'bitstring>=3.1.6',
            'aiodns>=2.0.0',
            'aioredis>=1.3.0',
            'aioredlock==0.3.0',
            'python-dotenv>=0.10.3',
            'python-rapidjson>=0.9.1',
            'nanopy>=20.0.0',
            'aiohttp>=3.6.2',
            'pyyaml>=5.3.1',
            'pycryptodome>=3.9.4',
            'aiounittest>=1.3.1',
            'websockets>=8.1',
            'setuptools'
        ]
        if sys.platform not in ('win32', 'cygwin', 'cli'):
            ret.append('uvloop>=0.14.0')
        return ret
setup(
    # Application name:
    name="pippin-wallet",
    # Version number:
    version=pippin.version.__version__,
    # Application author details:
    author="Appditto LLC",
    author_email="hello@appditto.com",
    # License
    license="MIT License",
    # Packages
    packages=find_packages(include=["pippin*"]),
    package_data={'pippin': ['*.yaml']},
    zip_safe=True,
    # Details
    url="https://github.com/appditto/pippin_nano_wallet",
    description="A production-ready, high-performance developer wallet for Nano and BANANO.",
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Framework :: AsyncIO",
        "Topic :: Security :: Cryptography",
        "Operating System :: POSIX",
        "Operating System :: MacOS :: MacOS X",
    ],
    keywords=(
        "cryptocurrency wallet nano banano "
        "bitcoin api aiohttp "
        "async asyncio aio"
    ),
    # Dependent packages (distributions)
    install_requires=requirements(),
    entry_points={
        'console_scripts': [
            'pippin-server = pippin.main:main',
            'pippin-cli = pippin.pippin_cli:main'
        ]
    }
)



from pathlib import Path, PurePath
import os

class Utils(object):
    """Generic utilities"""
    @staticmethod
    def get_project_root():
        home_path = Path.home()
        pippin_path = home_path.joinpath(PurePath('PippinData'))
        if not os.path.exists(pippin_path):
            os.makedirs(pippin_path)
        return pippin_path


from typing import List

import nanopy
import uuid
from aiohttp import log

from aioredlock import LockError
from tortoise import fields
from tortoise.functions import Max
from tortoise.models import Model
from tortoise.transactions import in_transaction

import pippin.db.models.account as acct
import pippin.db.models.adhoc_account as adhoc_acct
from pippin.db.redis import RedisDB
from pippin.model.secrets import SeedStorage
from pippin.util.crypt import AESCrypt
from pippin.util.wallet import WalletUtil
from pippin.network.rpc_client import AccountNotFound


class WalletNotFound(Exception):
    pass

class WalletLocked(Exception):
    def __init__(self, wallet):
        self.wallet = wallet
    pass

class AccountAlreadyExists(Exception):
    pass

class Wallet(Model):
    id = fields.UUIDField(pk=True)
    seed = fields.CharField(max_length=128, unique=True)
    representative = fields.CharField(max_length=65, null=True)
    encrypted = fields.BooleanField(default=False)
    work = fields.BooleanField(default=True)
    created_at = fields.DatetimeField(auto_now_add=True)

    class Meta:
        table = 'wallets'

    @staticmethod
    async def get_wallet(id: str) -> 'Wallet':
        """Get wallet with ID, raise WalletNotFound if not found, WalletLocked if encrypted"""
        wallet = await Wallet.filter(id=uuid.UUID(id)).first()
        if wallet is None:
            raise WalletNotFound()
        elif wallet.encrypted:
            decrypted = SeedStorage.instance().get_decrypted_seed(wallet.id)
            if decrypted is None:
                raise WalletLocked(wallet)
            wallet.seed = decrypted
        return wallet

    async def change_seed(self, seed: str):
        async with in_transaction() as conn:
            self.seed = seed
            for a in await self.accounts.all():
                a.address = nanopy.deterministic_key(self.seed, index=a.account_index)[2]
                await a.save(using_db=conn, update_fields=['address'])
            await self.save(using_db=conn, update_fields=['seed'])

    async def encrypt_wallet(self, password: str):
        """Encrypt wallet seed with password"""
        async with in_transaction() as conn:
            # If password is empty string then decrypted wallet
            if len(password.strip()) == 0:
                self.encrypted = False
                for a in await self.adhoc_accounts.all():
                    decrypted = SeedStorage.instnace().get_decrypted_seed(f"{self.id}:{a.address}")
                    if decrypted is not None:
                        a.private_key = decrypted
                    await a.save(using_db=conn, update_fields=['private_key'])
            else:
                crypt = AESCrypt(password)
                encrypted = crypt.encrypt(self.seed)
                self.seed = encrypted
                self.encrypted = True
                for a in await self.adhoc_accounts.all():
                    a.private_key = crypt.encrypt(a.private_key_get())
                    await a.save(using_db=conn, update_fields=['private_key'])            
            await self.save(using_db=conn, update_fields=['seed', 'encrypted'])

    async def unlock_wallet(self, password: str):
        """Unlock wallet with given password, raise DecryptionError if invalid password"""
        crypt = AESCrypt(password)
        decrypted = crypt.decrypt(self.seed)
        # Store decrypted wallet in memory
        SeedStorage.instance().set_decrypted_seed(self.id, decrypted)
        # Decrypt any ad-hoc accounts
        for a in await self.adhoc_accounts.all():
            SeedStorage.instance().set_decrypted_seed(f"{self.id}:{a.address}", crypt.decrypt(a.private_key))

    async def lock_wallet(self):
        """Lock wallet and remove all decrypted seeds from memory"""
        SeedStorage.instance().remove(self.id)
        for a in await self.adhoc_accounts.all():
            await SeedStorage.instance().remove(f"{self.id}:{a.address}")

    async def is_locked(self) -> bool:
        """Determine whether this wallet is locked or not"""
        return SeedStorage.instance().contains_encrypted(self.id) 

    async def bulk_representative_update(self, rep: str):
        """Set all account representatives to rep"""
        for a in await self.accounts.all():
            w = WalletUtil(a, self)
            try:
                await w.representative_set(rep, only_if_different=True)
            except AccountNotFound:
                pass

    async def adhoc_account_create(self, key: str, password: str = None) -> str:
        """Add an adhoc private key to the wallet, raise AccountAlreadyExists if it already exists"""
        pubkey = nanopy.ed25519_blake2b.publickey(bytes.fromhex(key)).hex()
        address = nanopy.account_get(pubkey)
        # See if address already exists
        a = await self.accounts.filter(address=address).first()
        if a is None:
            a = await self.adhoc_accounts.filter(address=address).first()
        if a is not None:
            raise AccountAlreadyExists(a)
        # Create it
        crypt = None
        if password is not None:
            crypt = AESCrypt(password)
        a = adhoc_acct.AdHocAccount(
            wallet=self,
            private_key=crypt.encrypt(key) if crypt is not None else key,
            address=address
        )
        await a.save()
        return address

    async def get_newest_account(self) -> acct.Account:
        """Get account with highest index beloning to this wallet"""
        return await acct.Account.filter(wallet=self).annotate(max_index=Max("account_index")).order_by('-account_index').first()

    async def account_create(self, using_db=None) -> str:
        """Create an account on this seed and return the created account"""
        async with await (await RedisDB.instance().get_lock_manager()).lock(f"pippin:{str(self.id)}:account_create") as lock:
            account = await self.get_newest_account()
            log.server_logger.debug(f"Creating account for {self.id}")
            index = account.max_index + 1 if account is not None and account.max_index is not None and account.max_index >= 0 else 0
            private_key, public_key, address = nanopy.deterministic_key(self.seed, index=index)
            log.server_logger.debug(f"Created {address} at index {index}")
            account = acct.Account(
                wallet=self,
                account_index=index,
                address=address
            )
            await account.save(using_db=using_db)
            return address

    async def accounts_create(self, count=0, using_db=None) -> List[str]:
        """Create {count} accounts on this seed and return the created accounts"""
        count = max(1, count)
        async with await (await RedisDB.instance().get_lock_manager()).lock(f"pippin:{str(self.id)}:account_create") as lock:
            account = await acct.Account.filter(wallet=self).annotate(max_index=Max("account_index")).order_by('-account_index').first()
            log.server_logger.debug(f"Creating {count} accounts for {self.id}")
            current_index = account.max_index + 1 if account is not None and account.max_index is not None and account.max_index >= 0 else 0
            accounts = []
            for i in range(count):
                private_key, public_key, address = nanopy.deterministic_key(self.seed, index=current_index)
                accounts.append(
                    acct.Account(
                        wallet=self,
                        account_index=current_index,
                        address=address
                    )
                )
                current_index+=1
            await acct.Account.bulk_create(
                accounts, using_db=using_db
            )

            return [a.address for a in accounts]

    async def get_account(self, address: str) -> acct.Account:
        """Get an an account that begins to this wallet"""
        a = await self.accounts.filter(address=address).first()
        if a is None:
            # Check adhoc
            a = await self.adhoc_accounts.filter(address=address).first()
        return a

    async def get_all_accounts(self) -> List[acct.Account]:
        """Get all accounts belong to this wallet"""
        accounts = await self.accounts.all()
        if accounts is None:
            accounts = []
        adhoc_accounts = await self.adhoc_accounts.all()
        if adhoc_accounts is not None:
            accounts += adhoc_accounts
        return accounts